from typing import Callable
from time import time

from sigmaepsilon.solid.fem.utils import irows_icols_bulk
from sigmaepsilon.solid.fem.mesh import FemMesh
from sigmaepsilon.solid.fem.imap import (index_mappers, box_spmatrix,
                                         box_rhs, box_dof_numbering)
from neumann.array import matrixform
from neumann.linalg.sparse.csr import csr_matrix as csr

from scipy.sparse import coo_matrix as npcoo, csc_matrix as npcsc
from scipy.sparse.linalg import spsolve
import numpy as np


from .utils import compliances_bulk, cells_around, \
    get_filter_factors, get_filter_factors_csr, weighted_stiffness_bulk, \
    filter_stiffness
from .filter import sensitivity_filter, sensitivity_filter_csr


def OC_SIMP_COMP(*args, sparsify=False, **kwargs):
    if (len(args) > 0) and isinstance(args[0], FemMesh):
        mesh = args[0]
        centers = mesh.centers()
        vols = mesh.areas()
        K_bulk = mesh.stiffness_matrix(sparse=False)
        gnum = mesh.element_dof_numbering()

        # mapping dofs
        loc_to_glob, glob_to_loc = index_mappers(gnum, return_inverse=True)
        gnum = box_dof_numbering(gnum, glob_to_loc)

        # essential boundary conditions
        Kp_coo = mesh.penalty_matrix_coo()
        Kp_coo = box_spmatrix(Kp_coo, glob_to_loc)

        # natural boundary conditions
        F = box_rhs(matrixform(mesh.load_vector()), loc_to_glob)
        if sparsify:
            F = npcsc(F)

        return _OC_SIMP_COMP_VOL_SENS_(K_bulk, Kp_coo, F, gnum, vols, centers, **kwargs)
    else:
        return _OC_SIMP_COMP_VOL_SENS_(*args, **kwargs)


def _OC_SIMP_COMP_VOL_SENS_(
        K_bulk: np.ndarray, Kp_coo: npcoo, RHS: np.ndarray,
        gnum: np.ndarray, vols: np.ndarray, centers: np.ndarray, *args,
        miniter=10, maxiter=100, p_start=1.0, p_stop=4.0,
        p_inc=0.2, p_step=5, q=0.5, vfrac=0.6, dtol=0.1,
        r_min=None, fltr_stiff=False, summary=True,
        callback: Callable = None, **kwargs):
    do_filter = r_min is not None

    # initial solution to set up parameters
    krows, kcols = irows_icols_bulk(gnum)
    krows = krows.flatten()
    kcols = kcols.flatten()
    nTOTV = gnum.max() + 1
    full_shape = (nTOTV, nTOTV)
    #
    dens = np.zeros_like(vols)
    dens_tmp = np.zeros_like(vols)
    dens_tmp_ = np.zeros_like(vols)
    dCdx = np.zeros_like(vols)

    # ------------------ INITIAL SOLUTION ---------------

    Kr_coo = npcoo((K_bulk.flatten(), (krows, kcols)),
                   shape=full_shape, dtype=K_bulk.dtype)
    K_coo = Kr_coo + Kp_coo
    U = spsolve(K_coo, RHS, permc_spec='NATURAL', use_umfpack=True)
    comps = compliances_bulk(K_bulk, U, gnum)
    np.clip(comps, 1e-7, None, out=comps)
    comp = np.sum(comps)
    vol = np.sum(vols)

    # initialization
    vol_start = vol
    vol_min = vfrac * vol_start
    if do_filter:
        neighbours = cells_around(centers, r_min, as_csr=False)
        if isinstance(neighbours, csr):
            factors = get_filter_factors_csr(centers, neighbours, r_min)
            fltr = sensitivity_filter_csr
        else:
            factors = get_filter_factors(centers, neighbours, r_min)
            fltr = sensitivity_filter

    # ------------- INITIAL FEASIBLE SOLUTION ------------

    dens[:] = vfrac
    K_bulk_w = weighted_stiffness_bulk(K_bulk, dens)
    Kr_coo = npcoo((K_bulk_w.flatten(), (krows, kcols)),
                   shape=full_shape, dtype=K_bulk.dtype)
    K_coo = Kr_coo + Kp_coo
    U = spsolve(K_coo, RHS, permc_spec='NATURAL', use_umfpack=True)
    comps[:] = compliances_bulk(K_bulk, U, gnum)
    np.clip(comps, 1e-7, None, out=comps)
    comp = np.sum(comps)
    if callback is not None:
        callback(0, comp, vol, dens)

    # ------------------- ITERATION -------------------

    p = p_start
    cIter = -1
    dt = 0
    terminate = False
    while not terminate:
        if (p < p_stop) and (np.mod(cIter, p_step) == 0):
            p += p_inc
        cIter += 1

        # estimate lagrangian
        lagr = p * comp / vol

        # set up boundaries of change
        _dens = dens * (1 - dtol)
        np.clip(_dens, 1e-5, 1.0, out=_dens)
        dens_ = dens * (1 + dtol)
        np.clip(dens_, 1e-5, 1.0, out=dens_)

        # sensitivity [*(-1)]
        dCdx[:] = p * comps * dens ** (p-1)

        # sensitivity filter
        if do_filter:
            dCdx[:] = fltr(dCdx, dens, neighbours, factors)

        # calculate new densities and lagrangian
        dens_tmp_[:] = dens * (dCdx / vols) ** q
        dens_tmp[:] = dens_tmp_
        _lagr = 0
        lagr_ = 2 * lagr
        while (lagr_ - _lagr) > 1e-3:
            _lagr_ = (_lagr + lagr_) / 2
            dens_tmp[:] = dens_tmp_ / (_lagr_ ** q)
            np.clip(dens_tmp, _dens, dens_, out=dens_tmp)
            vol_tmp = np.sum(dens_tmp * vols)
            if vol_tmp < vol_min:
                lagr_ = _lagr_
            else:
                _lagr = _lagr_
        lagr = lagr_
        dens[:] = dens_tmp

        # resolve equilibrium equations
        K_bulk_w = weighted_stiffness_bulk(K_bulk, dens)
        if fltr_stiff:
            # TODO : SELECT WORST <X> PERCENT
            Kr_coo = npcoo(filter_stiffness(K_bulk_w, gnum, dens, tol=0.001),
                           shape=full_shape, dtype=K_bulk.dtype)
        else:
            Kr_coo = npcoo((K_bulk_w.flatten(), (krows, kcols)),
                           shape=full_shape, dtype=K_bulk.dtype)
        K_coo = Kr_coo + Kp_coo
        t0 = time()
        U = spsolve(K_coo, RHS, permc_spec='NATURAL', use_umfpack=True)
        dt += time() - t0
        comps[:] = compliances_bulk(K_bulk, U, gnum)
        np.clip(comps, 1e-7, None, out=comps)
        comp = np.sum(comps)
        vol = np.sum(dens * vols)

        if callback is not None:
            callback(cIter, comp, vol, dens)

        if cIter < miniter:
            terminate = False
        elif cIter >= maxiter:
            terminate = True
        else:
            terminate = (p >= p_stop)

    if summary:
        dsum = {
            'avg. time [ms]': 1000 * dt / cIter,
            'niter': cIter,
            'filter_stiff': fltr_stiff
        }
        return dens, dsum
    return dens


if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
from linkeddeepdict import LinkedDeepDict

from .elem import FiniteElement
from .metaelem import ABCFiniteElement
from .celldata import CellData

from .bernoulli2 import Bernoulli2 as B2
from .bernoulli3 import Bernoulli3 as B3

from .cst import CSTM, CSTP
from .lst import LSTM, LSTP

from .q4 import Q4M, Q4P
from .q9 import Q9M, Q9P

from .tet4 import TET4
from .tet10 import TET10

from .h8 import H8
from .h27 import H27


cells_dict = LinkedDeepDict({
    'bernoulli' : {
        'B2' : B2, 'B3' : B3
    },
    'membrane' : {
        'T3' : CSTM, 'T6' : LSTM,
        'Q4' : Q4M, 'Q9' : Q9M,
    },
    'mindlin' : {
        'T3' : CSTP, 'T6' : LSTP,
        'Q4' : Q4P, 'Q9' : Q9P
    },
    '3d' : {
        'H8' : H8, 'H27' : H27, 
        'TET4' : TET4, 'TET10' : TET10
    }
})


# -*- coding: utf-8 -*-
import numpy as np
from numpy.linalg import norm
from scipy.sparse import isspmatrix as isspmatrix_np, spmatrix
from scipy.sparse.linalg import eigsh
from scipy.linalg import eigh


def calc_eig_res(A, M, eigvals, eigvecs) -> list:
    """
    Returns the residuals of the generalized eigenvalue problem
    
    .. math::
        :nowrap:

        \\begin{equation}
            \\mathbf{A} \\mathbf{x} = \\mathbf{v} \\mathbf{M} \\mathbf{x},
        \\end{equation}
    
    according to
    
    .. math::
        :nowrap:

        \\begin{equation}
            r_i = \\frac{|| (\\mathbf{A} - \\lambda_i \\mathbf{M}) ||}{|| \\mathbf{A} \\lambda_i ||},
        \\end{equation}
        
    Parameters
    ----------
    A : numpy.ndarray or scipy.linalg.sparse.spmatrix
            
    M : numpy.ndarray or scipy.linalg.sparse.spmatrix
    
    Returns
    -------
    list
        The residual for all eigenvectors.
        
    """
    def rfunc(i): return norm((A - eigvals[i] * M) @ eigvecs[:, i]) / \
            norm(A @ eigvecs[:, i])
    return list(map(rfunc, range(len(eigvals))))


def normalize_eigenvectors(vecs, A):
    """
    Returns the eigenvectors normalized to the matrix `A`.
    
    Parameters
    ----------
    vecs : numpy.ndarray
        The eigenvectors as a 2d array, where each column is
        an eigenvector.
            
    A : scipy.linalg.sparse.spmatrix or numpy.ndarray
        A 2d array.
    
    Returns
    -------
    numpy.ndarray
        The normalized eigenvectors.
        
    """
    N = vecs.shape[-1]
    def qprod(i): return vecs[:, i].T @ A @ vecs[:, i]
    foo, rng = lambda i: vecs[:, i] / qprod(i), range(N)
    return np.stack(list(map(foo, rng))).T


def eig_dense(A, *args, M:np.ndarray=None, normalize:bool=False, nmode:str='A',
              return_residuals:bool=False, **kwargs):
    """
    Returns all eigenvectors end eigenvalues for a dense square 
    matrix, for the standard eigenvalue problem
    
    .. math::
        :nowrap:

        \\begin{equation}
            \\mathbf{A} \\mathbf{x} = \\lambda \\mathbf{x},
        \\end{equation}
        
    or if :math:`\\mathbf{M}` is specified, the general eigenvalue problem

    .. math::
        :nowrap:

        \\begin{equation}
            \\mathbf{A} \\mathbf{x} = \\lambda \\mathbf{M} \\mathbf{x},
        \\end{equation}
    
    The values are calculated by calling :func:`scipy.linalg.eigs`.
    Extra keyword arguments are forwarded.

    Parameters
    ----------
    A : numpy.ndarray or scipy.linalg.sparse.spmatrix
            
    M : numpy.ndarray or scipy.linalg.sparse.spmatrix, Optional
    
    nmode : str, ['A' | 'M'], Optional
        Contrtols the normalization of the eigenvectors. Default is 'A'.

    normalize : bool, Optional
        Controls normalization of the eigenvectors. See the notes below. 
        Default is False.
    
    See also
    --------
    scipy.linalg.eigs

    """
    A_ = A.todense() if isspmatrix_np(A) else A
    if M is None:
        vals, vecs = eigh(A_, **kwargs)
    else:
        M_ = M.todense() if isspmatrix_np(M) else M
        vals, vecs = eigh(A_, b=M_, **kwargs)
    if normalize:
        if nmode == 'A':
            vecs = normalize_eigenvectors(vecs, A)
        elif nmode == 'M':
            vecs = normalize_eigenvectors(vecs, M)
        else:
            raise NotImplementedError()
    if return_residuals:
        r = calc_eig_res(A, M, vals, vecs)
        return vals, vecs, r
    return vals, vecs


def eig_sparse(A, *args, k:int=10, M:spmatrix=None, normalize:bool=False, which:str='SM',
               maxiter:int=None, nmode:str='A', return_residuals:bool=False, **kwargs):
    """
    Returns all eigenvectors end eigenvalues for a sparse square matrix,
    for the standard eigenvalue problem
    
    .. math::
        :nowrap:

        \\begin{equation}
            \\mathbf{A} \\mathbf{x} = \\lambda \\mathbf{x},
        \\end{equation}
        
    or if :math:`\\mathbf{M}` is specified, the general eigenvalue problem

    .. math::
        :nowrap:

        \\begin{equation}
            \\mathbf{A} \\mathbf{x} = \\lambda \\mathbf{M} \\mathbf{x},
        \\end{equation}

    The values are calculated by calling :func:`scipy.sparse.linalg.eigsh`, 
    which uses Arnoldi itrations. See references [1]_ and [2]_ for more details. 
    Extra keyword arguments are forwarded.

    Parameters
    ----------
    A : scipy.linalg.sparse.spmatrix
            
    M : scipy.linalg.sparse.spmatrix, Optional
    
    k : int or str, Optional
        Number of eigendata to calculate. If `k` is a string, it must be 'all'. 
        Default is 10.

    nmode : str, ['A' | 'M'], Optional
        Contrtols the normalization of the eigenvectors. Default is 'A'.

    normalize : bool, Optional
        Controls normalization of the eigenvectors. See the notes below. 
        Default is False.

    which : str, ['LM' | 'SM' | 'LR' | 'SR' | 'LI' | 'SI'], Optional
        Which `k` eigenvectors and eigenvalues to find:
            'LM' : largest magnitude
            'SM' : smallest magnitude
            'LR' : largest real part
            'SR' : smallest real part
            'LI' : largest imaginary part
            'SI' : smallest imaginary part
        Default is 'SM'.
    
    See also
    --------
    :func:`scipy.sparse.linalg.eigsh`

    References
    ----------
    .. [1] ARPACK Software, http://www.caam.rice.edu/software/ARPACK/
    
    .. [2] R. B. Lehoucq, D. C. Sorensen, and C. Yang, ARPACK USERS GUIDE:
           Solution of Large Scale Eigenvalue Problems by Implicitly Restarted
           Arnoldi Methods. SIAM, Philadelphia, PA, 1998.

    """
    vals, vecs = eigsh(A=A, k=k, M=M, which=which, maxiter=maxiter, **kwargs)
    if normalize:
        if nmode == 'A':
            vecs = normalize_eigenvectors(vecs, A)
        elif nmode == 'M':
            vecs = normalize_eigenvectors(vecs, M)
        else:
            raise NotImplementedError()
    if return_residuals:
        r = calc_eig_res(A, M, vals, vecs)
        return vals, vecs, r
    return vals, vecs



# -*- coding: utf-8 -*-
from .structure import Structure
from ..mesh import FemMesh


class Domain(Structure):
    
    def __init__(self, *args, mesh: FemMesh = None, **kwargs):
        super().__init__(wrap=mesh)
        
        
class VariableThicknessSheet(Domain):
    ...
        


# -*- coding: utf-8 -*-
import numpy as np
from numpy import sin, cos, ndarray, pi as PI
from numba import njit, prange
from typing import Iterable, Union
from neumann.array import (atleast1d, atleast2d, atleast3d,
                           atleast4d, itype_of_ftype)


__all__ = ['postproc', 'postproc_Mindlin', 'postproc_Kirchhoff']


UZ, ROTX, ROTY, CX, CY, CXY, EXZ, EYZ, MX, MY, MXY, QX, QY = list(range(13))


def postproc(size: Union[float, Iterable[float]], shape: Union[int, Iterable[int]],
             points: ndarray, solution: ndarray, loads: ndarray,
             D: Union[float, ndarray], S: Union[float, ndarray] = None, *,
             squeeze: bool = True):
    """
    Calculates postprocessing items.
    """
    #ABDS = atleast3d(ABDS)
    ftype = solution.dtype
    itype = itype_of_ftype(ftype)
    if S is not None:
        if not isinstance(size, Iterable):
            res = postproc_Timoshenko(size, shape,
                                      atleast1d(points).astype(ftype),
                                      atleast3d(solution).astype(ftype),
                                      atleast1d(D), atleast1d(S))
        else:
            res = postproc_Mindlin(np.array(size).astype(ftype),
                                   np.array(shape).astype(itype),
                                   atleast2d(points).astype(ftype),
                                   atleast4d(solution).astype(ftype),
                                   atleast3d(D), atleast3d(S))
    else:
        if not isinstance(size, Iterable):
            res = postproc_Bernoulli(size, shape,
                                     atleast1d(points).astype(ftype),
                                     atleast3d(solution).astype(ftype),
                                     atleast1d(D), loads)
        else:
            res = postproc_Kirchhoff(np.array(size).astype(ftype),
                                     np.array(shape).astype(itype),
                                     atleast2d(points).astype(ftype),
                                     atleast3d(solution).astype(ftype),
                                     atleast3d(D), loads)
    # (N, nRHS, nLHS, nP, ...)
    res = np.sum(res, axis=0)
    # (nRHS, nLHS, nP, ...)
    return res if not squeeze else np.squeeze(res)


@njit(nogil=True, parallel=True, cache=True)
def postproc_Bernoulli(L: float, N: int, points: ndarray,
                       solution: ndarray, EI: ndarray, loads):
    """
    JIT-compiled function that calculates post-processing quantities 
    at selected ponts for multiple left- and right-hand sides.       

    Parameters
    ----------
    L : float
        The length of the beam.

    N : int 
        The number of harmonic terms.

    points : numpy.ndarray 
        1d array of shape (nP,) of coordinates.

    solution : numpy.ndarray 
        Results of a Navier solution as a 3d array of shape (nRHS, nLHS, N).

    EI : float 
        Bending stiffnesses as a 3d float array of shape (nLHS, 3, 3).

    Returns
    -------
    numpy.ndarray 
        5d array of shape (N, nRHS, nLHS, nP, ...) of post-processing items. 
        The indices along the last axis denote the following quantities:

            0 : displacement

            1 : rotation

            2 : curvature

            3 : shear strain

            4 : moment

            5 : shear force

    """
    nP = points.shape[0]
    nRHS, nLHS = solution.shape[:2]
    res = np.zeros((N, nRHS, nLHS, nP, 6), dtype=solution.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            for iP in prange(nP):
                x = points[iP]
                for n in prange(1, N + 1):
                    iN = n - 1
                    arg = PI*n/L
                    Sn = sin(x*arg)
                    Cn = cos(x*arg)
                    vn = solution[iRHS, iLHS, iN]
                    q = loads[iRHS, iN, 1]
                    res[iN, iRHS, iLHS, iP, 0] = vn * Sn
                    res[iN, iRHS, iLHS, iP, 1] = vn * arg * Cn
                    res[iN, iRHS, iLHS, iP, 2] = - vn * arg**2 * Sn
                    res[iN, iRHS, iLHS, iP, 4] = - EI[iLHS] * vn * arg**2 * Sn
                    res[iN, iRHS, iLHS, iP, 5] = (EI[iLHS] * vn * arg**3 + q) * Cn
    return res


@njit(nogil=True, parallel=True, cache=True)
def postproc_Timoshenko(L: float, N: int, points: ndarray,
                        solution: ndarray, EI: ndarray, GA: ndarray):
    """
    JIT-compiled function that calculates post-processing quantities 
    at selected ponts for multiple left- and right-hand sides.       

    Parameters
    ----------
    L : float
        The length of the beam.

    N : int 
        The number of harmonic terms.

    points : numpy.ndarray 
        1d array of shape (nP,) of coordinates.

    solution : numpy.ndarray 
        Results of a Navier solution as a 4d array of shape (nRHS, nLHS, N, 2).

    EI : float 
        Bending stiffnesses as an 1d float array of shape (nLHS).
        
    GA : float 
        Corrected shear stiffnesses as an 1d float array of shape (nLHS).

    Returns
    -------
    numpy.ndarray 
        5d array of shape (N, nRHS, nLHS, nP, ...) of post-processing items. 
        The indices along the last axis denote the following quantities:

            0 : displacement

            1 : rotation

            2 : curvature

            3 : shear strain

            4 : moment

            5 : shear force

    """
    nP = points.shape[0]
    nRHS, nLHS = solution.shape[:2]
    res = np.zeros((N, nRHS, nLHS, nP, 6), dtype=solution.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            for iP in prange(nP):
                x = points[iP]
                for n in prange(1, N + 1):
                    iN = n - 1
                    arg = PI * n / L
                    Sn = sin(x * arg)
                    Cn = cos(x * arg)
                    vn, rn = solution[iRHS, iLHS, iN]
                    res[iN, iRHS, iLHS, iP, 0] = vn * Sn
                    res[iN, iRHS, iLHS, iP, 1] = rn * Cn
                    res[iN, iRHS, iLHS, iP, 2] = - rn * arg * Sn
                    res[iN, iRHS, iLHS, iP, 3] = (vn * arg - rn) * Cn
                    res[iN, iRHS, iLHS, iP, 4] = - EI[iLHS] * rn * arg * Sn
                    res[iN, iRHS, iLHS, iP, 5] = GA[iLHS] * (vn * arg - rn) * Cn
    return res


@njit(nogil=True, parallel=True, cache=True)
def postproc_Mindlin(size, shape: ndarray, points: ndarray,
                     solution: ndarray, D: ndarray, S: ndarray):
    """
    JIT-compiled function that calculates post-processing quantities 
    at selected ponts for multiple left- and right-hand sides.       

    Parameters
    ----------
    size : numpy.ndarray
        Sizes in both directions as an 1d float array of length 2.

    shape : numpy.ndarray 
        Number of harmonic terms involved in both directions as an
        1d integer array of length 2.

    points : numpy.ndarray 
        2d array of shape (nP, 2) of coordinates.

    solution : numpy.ndarray 
        Results of a Navier solution as a 4d array of shape (nRHS, nLHS, M * N, 3).

    D : numpy.ndarray 
        Bending stiffnesses as a 3d float array of shape (nLHS, 3, 3).

    S : numpy.ndarray
        Corrected shear stiffness as a 3d float array of shape (nLHS, 2, 2).

    Returns
    -------
    numpy.ndarray 
        5d array of shape (M * N, nRHS, nLHS, nP, ...) of post-processing items. 
        The indices along the last axis denote the following quantities:

            0 : displacement z

            1 : rotation x

            2 : rotation y

            3 : curvature x

            4 : curvature y

            5 : curvature xy

            6 : shear strain xz

            7 : shear strain yz

            8 : moment y

            9 : moment y

            10 : moment xy

            11 : shear force x

            12 : shear force y

    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nRHS, nLHS = solution.shape[:2]
    res = np.zeros((M * N, nRHS, nLHS, nP, 13), dtype=D.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
                D[iLHS, 1, 1], D[iLHS, 2, 2]
            S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
            for iP in prange(nP):
                xp, yp = points[iP, :2]
                for m in prange(1, M + 1):
                    Sm = sin(PI*m*xp/Lx)
                    Cm = cos(PI*m*xp/Lx)
                    for n in prange(1, N + 1):
                        iMN = (m-1) * N + n - 1
                        Amn, Bmn, Cmn = solution[iRHS, iLHS, iMN]
                        Sn = sin(PI*n*yp/Ly)
                        Cn = cos(PI*n*yp/Ly)
                        res[iMN, iRHS, iLHS, iP, UZ] = Cmn * Sm*Sn
                        res[iMN, iRHS, iLHS, iP, ROTX] = Amn * Sm*Cn
                        res[iMN, iRHS, iLHS, iP, ROTY] = Bmn * Sn*Cm
                        res[iMN, iRHS, iLHS, iP, CX] = -PI*Bmn * m*Sm*Sn/Lx
                        res[iMN, iRHS, iLHS, iP, CY] = PI*Amn * n*Sm*Sn/Ly
                        res[iMN, iRHS, iLHS, iP, CXY] = -PI*Amn*m*Cm*Cn/Lx + \
                            PI*Bmn*n*Cm*Cn/Ly
                        res[iMN, iRHS, iLHS, iP, EXZ] = Bmn*Sn*Cm + \
                            PI*Cmn*m*Sn*Cm/Lx
                        res[iMN, iRHS, iLHS, iP, EYZ] = -Amn*Sm*Cn + \
                            PI*Cmn*n*Sm*Cn/Ly
                        res[iMN, iRHS, iLHS, iP, MX] = PI*Amn*D12*n*Sm*Sn/Ly - \
                            PI*Bmn*D11*m*Sm*Sn/Lx
                        res[iMN, iRHS, iLHS, iP, MY] = PI*Amn*D22*n*Sm*Sn/Ly - \
                            PI*Bmn*D12*m*Sm*Sn/Lx
                        res[iMN, iRHS, iLHS, iP, MXY] = -PI*Amn*D66*m*Cm*Cn/Lx + \
                            PI*Bmn*D66*n*Cm*Cn/Ly
                        res[iMN, iRHS, iLHS, iP, QX] = Bmn*S55*Sn*Cm + \
                            PI*Cmn*S55*m*Sn*Cm/Lx
                        res[iMN, iRHS, iLHS, iP, QY] = -Amn*S44*Sm*Cn + \
                            PI*Cmn*S44*n*Sm*Cn/Ly
    return res


@njit(nogil=True, parallel=True, cache=True)
def postproc_Kirchhoff(size, shape: ndarray, points: ndarray,
                       solution: ndarray, D: ndarray, loads: ndarray):
    """
    JIT-compiled function that calculates post-processing quantities 
    at selected ponts for multiple left- and right-hand sides.       

    Parameters
    ----------
    size : numpy.ndarray
        Sizes in both directions as an 1d float array of length 2.

    shape : numpy.ndarray 
        Number of harmonic terms involved in both directions as an
        1d integer array of length 2.

    points : numpy.ndarray 
        2d array of point coordinates of shape (nP, 2).

    solution : numpy.ndarray 
        results of a Navier solution as a 3d array of shape (nRHS, nLHS, M * N).

    D : numpy.ndarray 
        3d array of bending stiffness terms (nLHS, 3, 3).

    Returns
    -------
    numpy.ndarray[M * N, nRHS, nLHS, nP, ...] 
        numpy array of post-processing items. The indices along 
        the last axpis denote the following quantities:

            0 : displacement z

            1 : rotation x

            2 : rotation y

            3 : curvature x

            4 : curvature y

            5 : curvature xy

            6 : shear strain xz

            7 : shear strain yz

            8 : moment y

            9 : moment y

            10 : moment xy

            11 : shear force x

            12 : shear force y

    """
    Lx, Ly = size
    M, N = shape
    nP = points.shape[0]
    nRHS, nLHS = solution.shape[:2]
    res = np.zeros((M * N, nRHS, nLHS, nP, 13), dtype=D.dtype)
    for iRHS in prange(nRHS):
        for iLHS in prange(nLHS):
            D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
                D[iLHS, 1, 1], D[iLHS, 2, 2]
            for iP in prange(nP):
                x, y = points[iP, :2]
                for m in prange(1, M + 1):
                    Sm = sin(PI*m*x/Lx)
                    Cm = cos(PI*m*x/Lx)
                    for n in prange(1, N + 1):
                        Sn = sin(PI*n*y/Ly)
                        Cn = cos(PI*n*y/Ly)
                        iMN = (m-1) * N + n - 1
                        Cmn = solution[iRHS, iLHS, iMN]
                        qxx, qyy = loads[iRHS, iMN, :2]
                        res[iMN, iRHS, iLHS, iP, UZ] = Cmn * Sm*Sn
                        res[iMN, iRHS, iLHS, iP, ROTX] = PI * Cmn*n*Sm*Cn/Ly
                        res[iMN, iRHS, iLHS, iP, ROTY] = -PI * Cmn*m*Sn*Cm/Lx
                        res[iMN, iRHS, iLHS, iP, CX] = PI**2 * \
                            Cmn * m**2*Sm*Sn/Lx**2
                        res[iMN, iRHS, iLHS, iP, CY] = PI**2 * \
                            Cmn * n**2*Sm*Sn/Ly**2
                        res[iMN, iRHS, iLHS, iP, CXY] = - \
                            2*PI**2 * Cmn*m*n*Cm*Cn/(Lx*Ly)
                        res[iMN, iRHS, iLHS, iP, MX] = PI**2*Cmn*D11*m**2*Sm*Sn/Lx**2 + \
                            PI**2*Cmn*D12*n**2 * Sm*Sn/Ly**2
                        res[iMN, iRHS, iLHS, iP, MY] = PI**2*Cmn*D12*m**2*Sm*Sn/Lx**2 + \
                            PI**2*Cmn*D22*n**2 * Sm*Sn/Ly**2
                        res[iMN, iRHS, iLHS, iP, MXY] = - \
                            2*PI**2*Cmn * D66*m*n*Cm*Cn/(Lx*Ly)
                        res[iMN, iRHS, iLHS, iP, QX] = PI**3*Cmn*D11*m**3*Sn*Cm/Lx**3 + \
                            PI**3*Cmn*D12*m*n**2*Sn*Cm/(Lx*Ly**2) + \
                            2*PI**3*Cmn*D66*m*n**2*Sn*Cm/(Lx*Ly**2) + qxx*Sn*Cm
                        res[iMN, iRHS, iLHS, iP, QY] = PI**3*Cmn*D12*m**2*n*Sm*Cn/(Lx**2*Ly) + \
                            PI**3*Cmn*D22*n**3*Sm*Cn/Ly**3 + \
                            2*PI**3*Cmn*D66*m**2*n*Sm*Cn/(Lx**2*Ly) + qyy*Sm*Cn
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange
__cache = True


@njit(nogil=True, cache=__cache)
def shape_function_values(x: float, L: float = 2):
    """
    Evaluates the shape functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            0.5 - 0.5 * x,
            0.25 * x**3 - 0.75 * x + 0.5,
            0.25 * x**3 - 0.75 * x + 0.5,
            0.5 - 0.5 * x,
            -0.125 * L * (x - 1)**2 * (x + 1),
            0.125 * L * (x - 1)**2 * (x + 1)
        ],
        [
            0.5 * x + 0.5,
            -0.25 * x**3 + 0.75 * x + 0.5,
            -0.25 * x**3 + 0.75 * x + 0.5,
            0.5 * x + 0.5,
            -0.125 * L * (x - 1) * (x + 1)**2,
            0.125 * L * (x - 1) * (x + 1)**2
        ]
    ])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_1(x: float, L: float = 2):
    """
    Evaluates the first derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            -0.500000000000000,
            0.75 * x**2 - 0.75,
            0.75 * x**2 - 0.75,
            -0.500000000000000,
            L * (-0.375 * x**2 + 0.25 * x + 0.125),
            0.375 * L * (x - 1) * (1.0 * x + 0.333333333333333)
        ],
        [
            0.500000000000000,
            0.75 - 0.75 * x**2,
            0.75 - 0.75 * x**2,
            0.500000000000000,
            L * (-0.375 * x**2 - 0.25 * x + 0.125),
            0.375 * L * (1.0 * x - 0.333333333333333) * (x + 1)
        ]
    ])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_2(x: float, L: float = 2):
    """
    Evaluates the second derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            0,
            1.5 * x,
            1.5 * x,
            0,
            L * (0.25 - 0.75 * x),
            L * (0.75 * x - 0.25)
        ],
        [
            0,
            -1.5 * x,
            -1.5 * x,
            0,
            L * (-0.75 * x - 0.25),
            L * (0.75 * x + 0.25)
        ]
    ])


@njit(nogil=True, cache=__cache)
def shape_function_derivatives_3(x: float, L: float = 2):
    """
    Evaluates the third derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        A 2d float array.

    """
    return np.array([
        [
            0,
            1.50000000000000,
            1.50000000000000,
            0,
            -0.75 * L,
            0.75 * L
        ],
        [
            0,
            -1.50000000000000,
            -1.50000000000000,
            0,
            -0.75 * L,
            0.75 * L
        ]
    ])


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_values_bulk(x: ndarray, L: ndarray):
    """
    Evaluates the shape functions at several points
    in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].

    L : numpy.ndarray
        Lengths of the beam elements. Default is None.

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nP, nNE=2, nDOF=6).
    """
    nP = x.shape[0]
    nE = L.shape[0]
    res = np.zeros((nE, nP, 2, 6), dtype=x.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_values(x[iP], L[iE])
    return res


@njit(nogil=True, cache=__cache)
def shape_function_derivatives(x: float, L: float = 2):
    """
    Evaluates the derivatives of the shape
    functions at a point x in the range [-1, 1].

    Parameters
    ----------
    x : float
        A point in the range [-1, 1].

    L : float, Optional
        Length of the beam element. Default is 2.

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nNE=2, nDOF=6, 3).
    """
    res = np.zeros((2, 6, 3))
    res[:, :, 0] = shape_function_derivatives_1(x, L)
    res[:, :, 1] = shape_function_derivatives_2(x, L)
    res[:, :, 2] = shape_function_derivatives_3(x, L)
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def shape_function_derivatives_bulk(x: ndarray, L: ndarray):
    """
    Evaluates the derivatives of the shape
    functions at several points in the range [-1, 1].

    Parameters
    ----------
    x : numpy.ndarray
        1d array of floats in the range [-1, -1].

    L : numpy.ndarray
        Lengths of the beam elements. Default is None.

    Returns
    -------
    numpy.ndarray
        5d float array of shape (nE, nP, nNE=2, nDOF=6, 3).
    """
    nP = x.shape[0]
    nE = L.shape[0]
    res = np.zeros((nE, nP, 2, 6, 3), dtype=x.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_derivatives(x[iP], L[iE])
    return res



# -*- coding: utf-8 -*-
import json
import numpy as np
from numpy import ndarray
from typing import Union, Iterable, Any

from dewloosh.core.tools import (allinkwargs, popfromkwargs, 
                                 float_to_str_sig)
from linkeddeepdict import LinkedDeepDict
from linkeddeepdict.tools.dtk import parsedicts_addr

from .problem import NavierProblem
from .preproc import (rhs_rect_const, rhs_conc_1d, rhs_conc_2d, 
                      rhs_line_const)


class NavierLoadError(Exception):
    """
    Exception raised for invalid load inputs.
    """

    def __init__(self, message=None):
        if message is None:
            message = ("Invalid input for loads. "
                        "It must be a dictionary, a NumPy array, or "
                        "an instance of `sigmaepsilon.solid.fourier.LoadGroup`.")
        super().__init__(message)


class LoadGroup(LinkedDeepDict):
    """
    A class to handle load groups for Navier's semi-analytic solution of
    rectangular plates and beams with specific boundary conditions.
    
    This class is also the base class of all other load types.
    
    See Also
    --------
    :class:`LinkedDeepDict`
    
    Examples
    --------
    
    >>> from sigmaepsilon.solid.fourier import LoadGroup, PointLoad
    >>> loads = LoadGroup(   
    >>>     group1 = LoadGroup(
    >>>         case1 = PointLoad(x=L/3, v=[1.0, 0.0]),
    >>>         case2 = PointLoad(x=L/3, v=[0.0, 1.0]),
    >>>     ),
    >>>     group2 = LoadGroup(
    >>>         case1 = PointLoad(x=2*L/3, v=[1.0, 0.0]),
    >>>         case2 = PointLoad(x=2*L/3, v=[0.0, 1.0]),
    >>>     ),    
    >>> )
    
    Since the LoadGroup is a subclass of LinkedDeepDict,
    a case is accessible as
    
    >>> loads['group1', 'case1']
    
    If you want to protect the object from the accidental
    creation of nested subdirectories, you can lock the layout
    by typing
    
    >>> loads.lock()
         
    """
    _typestr_ = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__problem = None

    @property
    def problem(self) -> NavierProblem:
        """
        Returns the attached problem.
        """
        if self.is_root():
            return self.__problem
        return self.root().problem
    
    @problem.setter
    def problem(self, value:NavierProblem):
        """
        Sets the attached problem.
        
        Parameters
        ----------
        value : NavierProblem
            The problem the loads are defined for.
        """
        assert self.is_root(), "The problem can only be set on the top-level object."
        self.__problem = value

    def blocks(self, *args, inclusive:bool=False, blocktype:Any=None,
               deep:bool=True, **kwargs) -> Iterable['LoadGroup']:
        """
        Returns a generator object that yields all the subgroups.
        
        Parameters
        ----------
        inclusive : bool, Optional
            If True, returns the object the call is made upon.
            Default is False.
            
        blocktype : Any, Optional
            The type of the load groups to return. Default is None, that
            returns all types.
            
        deep : bool, Optional
            If True, all deep groups are returned separately. Default is True.
        
        Yields
        ------
        LoadGroup
        
        """
        dtype = LoadGroup if blocktype is None else blocktype
        return self.containers(self, inclusive=inclusive,
                               dtype=dtype, deep=deep)

    def cases(self, *args, inclusive:bool=True, **kwargs) -> Iterable['LoadGroup']:
        """
        Returns a generator that yields the load cases in the group.
        
        Parameters
        ----------
        inclusive : bool, Optional
            If True, returns the object the call is made upon.
            Default is True.
            
        blocktype : Any, Optional
            The type of the load groups to return. Default is None, that
            returns all types.
            
        deep : bool, Optional
            If True, all deep groups are returned separately. Default is True.
        
        Yields
        ------
        LoadGroup
        
        """
        return filter(lambda i: i.__class__._typestr_ is not None, 
                      self.blocks(*args, inclusive=inclusive, **kwargs))

    @staticmethod
    def _string_to_dtype_(string: str = None):
        if string == 'group':
            return LoadGroup
        elif string == 'rectangle':
            return RectangleLoad
        elif string == 'point':
            return PointLoad
        elif string == 'line':
            return LineLoad
        else:
            raise NotImplementedError()

    def dump(self, path:str, *, mode:str='w', indent:int=4):
        """
        Dumps the content of the object to a file.
        
        Parameters
        ----------
        path : str
            The path of the file on your filesystem.
            
        mode : str, Optional
            https://www.programiz.com/python-programming/file-operation
            
        indent : int, Optional
            Governs the level to which members will be pretty-printed.
            Default is 4.

        """
        with open(path, mode) as f:
            json.dump(self.to_dict(), f, indent=indent)

    @classmethod
    def from_json(cls, path: str = None) -> 'LoadGroup':
        """
        Loads a loadgroup from a JSON file.
        
        Parameters
        ----------
        path : str
            The path to a file on your filesystem.
            
        Returns
        -------
        LoadGroup
        
        """
        if path is not None:
            with open(path, 'r') as f:
                d = json.load(f)
            return cls.from_dict(d)

    def _encode_(self, *args, **kwargs) -> dict:
        """
        Returns the group as a dictionary. Overwrite this in child 
        implementations.
        """
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
        }
        return res

    @classmethod
    def _decode_(cls, d: dict = None, *args, **kwargs):
        """
        Returns a LoadGroup object from a dictionary. 
        Overwrite this in child implementations.
        
        Parameters
        ----------
        d : dict, Optional
            A dictionary.
        
        **kwargs : dict, Optional
            Keyword arguments defining a load group.
            
        """
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            'key': d.pop('key', None),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def to_dict(self) -> dict:
        """
        Returns the group as a dictionary.
        """
        res = self._encode_()
        for key, value in self.items():
            if isinstance(value, LoadGroup):
                res[key] = value.to_dict()
            else:
                res[key] = value
        return res

    @staticmethod
    def from_dict(d: dict = None, **kwargs) -> 'LoadGroup':
        """
        Reads a LoadGroup from a dictionary. The keys in the dictionaries
        must match the parameters of the corresponding load types and a type
        indicator.
        """
        if 'type' in d:
            cls = LoadGroup._string_to_dtype_(d['type'])
        else:  
            cls = LoadGroup
        res = cls(**d)
        for addr, value in parsedicts_addr(d, inclusive=True):
            if len(addr) == 0:
                continue
            if 'type' in value:
                cls = LoadGroup._string_to_dtype_(value['type'])
            else: 
                cls = LoadGroup
            value['key'] = addr[-1]
            res[addr] = cls(**value)
        return res

    def __repr__(self):
        return 'LoadGroup(%s)' % (dict.__repr__(self))


class RectangleLoad(LoadGroup):
    """
    A class to handle rectangular loads.
    
    Parameters
    ----------
    value: Iterable
        1d or 2d iterable of scalars for all 3 degrees of
        freedom in the order :math:`mx, my, fz`.
        
    points: Iterable, Optional
        The coordinates of the lower-left and upper-right points of the region
        where the load is applied. Default is None.

    **kwargs : dict, Optional
        If the region of application is not specified by the argument 'points',
        extra keyword arguments are forwarded to :func:`get_coords`. Default is None.
    
    """
    _typestr_ = 'rectangle'

    def __init__(self, *args, value:Iterable, points:Iterable=None, **kwargs):
        if points is not None:
            self.points = np.array(points, dtype=float)
        else:
            self.points = RectangleLoad.get_coords(kwargs)
        self.value = np.array(value, dtype=float)
        super().__init__()

    def _encode_(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
            'region': float_to_str_sig(self.region(), sig=6),
            'value': float_to_str_sig(self.value, sig=6),
        }
        return res

    @classmethod
    def _decode_(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        points = RectangleLoad.get_coords(d)
        clskwargs = {
            'key': d.pop('key', None),
            'points': points,
            'value': np.array(d.pop('value'), dtype=float)
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    @staticmethod
    def get_coords(d: dict = None, *args, **kwargs):
        """
        Returns the bottom-left and upper-right coordinates of the region
        of the load from several inputs.
        
        Parameters
        ----------
        d : dict, Optional
            A dictionary, which is equivalrent to a parameter set from the
            other parameters listed here. Default is None.
            
        region : Iterable, Optional
            An iterable of length 4 with values x0, y0, w, and h. Here x0 and y0 are
            the coordinates of the bottom-left corner, w and h are the width and height
            of the region.
            
        xy : Iterable, Optional
            The position of the bottom-left corner as an iterable of length 2.
            
        w : float, Optional
            The width of the region.
            
        h : float, Optional
            The height of the region.
            
        center : Iterable, Optional
            The coordinates of the center of the region.
            
        Returns
        -------
        numpy.ndarray
            A 2d float array of coordinates, where the entries of the first and second 
            rows are the coordinates of the lower-left and upper-right points of the region.
        
        Examples
        --------
        The following definitions return the same output:
        
        >>> from sigmaepsilon.solid.fourier import RectLoad
        >>> RectLoad.get_coords(region=[2, 3, 0.5, 0.7])
        >>> RectLoad.get_coords(xy=[2, 3], w=0.5, h=0.7)
        >>> RectLoad.get_coords(center=[2.25, 3.35], w=0.5, h=0.7)
        >>> RectLoad.get_coords(dict(center=[2.25, 3.35], w=0.5, h=0.7))
        
        """
        points = None
        if d is None:
            d = kwargs
        try:
            if 'points' in d:
                points = np.array(d.pop('points'))
            elif 'region' in d:
                x0, y0, w, h = np.array(d.pop('region'))
                points = np.array([[x0, y0], [x0 + w, y0 + h]])
            elif allinkwargs(['xy', 'w', 'h'], **d):
                (x0, y0), w, h = popfromkwargs(['xy', 'w', 'h'], d)
                points = np.array([[x0, y0], [x0 + w, y0 + h]])
            elif allinkwargs(['center', 'w', 'h'], **d):
                (xc, yc), w, h = popfromkwargs(['center', 'w', 'h'], d)
                points = np.array([[xc - w/2, yc - h/2],
                                   [xc + w/2, yc + h/2]])
        except Exception as e:
            print(e)
            return None
        return points

    def region(self) -> Iterable:
        """
        Returns the region as a list of 4 values x0, y0, w, and h, where x0 and y0 are
        the coordinates of the bottom-left corner, w and h are the width and height
        of the region.
        """
        assert self.points is not None, "There are no points defined."
        return _points_to_region_(self.points)

    def rhs(self, *, problem:NavierProblem=None) -> ndarray:
        """
        Returns the coefficients as a NumPy array.
        
        Parameters
        ----------
        problem : NavierProblem, Optional
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.
            
        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number 
            of harmonic terms involved (defined for the problem).
            
        """
        p = problem if problem is not None else self.problem
        return rhs_rect_const(p.size, p.shape, self.value, self.points)

    def __repr__(self):
        return 'RectangleLoad(%s)' % (dict.__repr__(self))


class LineLoad(LoadGroup):
    """
    A class to handle loads over lines.
    
    Parameters
    ----------
    x : Iterable
        The point of application as an 1d iterable for a beam, a 2d iterable
        for a plate. In the latter case, the first row is the first point, the
        second row is the second point.
        
    v : Iterable
        Load intensities for each dof. The order of the dofs for a beam
        is [F, M], for a plate it is [F, Mx, My].
        
    """
    _typestr_ = 'line'

    def __init__(self, *args, x:Iterable=None, 
                 v:Iterable=None, **kwargs):
        super().__init__(*args, x=x, v=v, **kwargs)
        
    @classmethod
    def _decode_(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            'key': d.pop('key', None),
            'x': np.array(d.pop('x')),
            'v': np.array(d.pop('v')),
        }
        clskwargs.update(d)
        return cls(**clskwargs)
    
    def _encode_(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
            'x': float_to_str_sig(self['x'], sig=6),
            'v': float_to_str_sig(self['v'], sig=6),
        }
        return res
        
    def rhs(self, *, problem:NavierProblem=None) -> ndarray:
        """
        Returns the coefficients as a NumPy array.
        
        Parameters
        ----------
        problem : NavierProblem, Optional
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.
            
        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number 
            of harmonic terms involved (defined for the problem).
            
        """
        p = problem if problem is not None else self.problem
        x = np.array(self['x'], dtype=float) 
        v = np.array(self['v'], dtype=float)
        return rhs_line_const(p.length, p.N, x, v)

    def __repr__(self):
        return 'LineLoad(%s)' % (dict.__repr__(self))


class PointLoad(LoadGroup):
    """
    A class to handle concentrated loads.
    
    Parameters
    ----------
    x : Union[float, Iterable]
        The point of application. A scalar for a beam, an iterable of
        length 2 for a plate.
        
    v : Iterable
        Load values for each dof. The order of the dofs for a beam
        is [F, M], for a plate it is [F, Mx, My].
    
    """
    _typestr_ = 'point'

    def __init__(self, *args, x:Union[float, Iterable]=None, 
                 v:Iterable=None, **kwargs):
        super().__init__(*args, x=x, v=v, **kwargs)

    @classmethod
    def _decode_(cls, d: dict = None, *args, **kwargs):
        if d is None:
            d = kwargs
            kwargs = None
        if kwargs is not None:
            d.update(kwargs)
        clskwargs = {
            'key': d.pop('key', None),
            'x': np.array(d.pop('x')),
            'v': np.array(d.pop('v')),
        }
        clskwargs.update(d)
        return cls(**clskwargs)

    def _encode_(self, *args, **kwargs) -> dict:
        res = {}
        cls = type(self)
        res = {
            'type': cls._typestr_,
            'key': self.key,
            'x': float_to_str_sig(self['x'], sig=6),
            'v': float_to_str_sig(self['v'], sig=6),
        }
        return res
    
    def rhs(self, *, problem:NavierProblem=None) -> ndarray:
        """
        Returns the coefficients as a NumPy array.
        
        Parameters
        ----------
        problem : NavierProblem, Optional
            A problem the coefficients are generated for. If not specified,
            the attached problem of the object is used. Default is None.
            
        Returns
        -------
        numpy.ndarray
            2d float array of shape (H, 3), where H is the total number 
            of harmonic terms involved (defined for the problem).
            
        """
        p = problem if problem is not None else self.problem
        x = self['x'] 
        v = np.array(self['v'])
        if hasattr(p, 'size'):
            return rhs_conc_2d(p.size, p.shape, v, x)
        else:
            return rhs_conc_1d(p.length, p.N, v, x)

    def __repr__(self):
        return 'PointLoad(%s)' % (dict.__repr__(self))


def _points_to_region_(points: ndarray):
    xmin = points[:, 0].min()
    ymin = points[:, 1].min()
    xmax = points[:, 0].max()
    ymax = points[:, 1].max()
    return xmin, ymin, xmax - xmin, ymax - ymin


# -*- coding: utf-8 -*-
from typing import Any, Iterable

import numpy as np
from numpy import ndarray
from sectionproperties.analysis.section import Section

from polymesh.config import __hasplotly__, __hasmatplotlib__
if __hasplotly__:
    from dewloosh.plotly import plot_lines_3d

from ..model.bernoulli.section import BeamSection
from .mesh import FemMesh
from .cells import B2, B3


class LineMesh(FemMesh):
    """
    A data class dedicated to 1d cells. It handles sections and other line 
    related information, plotting, etc.
    
    See Also
    --------
    :class:`sigmaepsilon.solid.fem.mesh.FemMesh`
        
    """

    _cell_classes_ = {
        2: B2,
        3: B3,
    }

    def __init__(self, *args, areas:ndarray=None, model:ndarray=None, 
                 section:BeamSection=None, **kwargs):

        if section is None:
            if isinstance(model, Section):
                section = BeamSection(wrap=model)
        if isinstance(section, BeamSection):
            section.calculate_geometric_properties()
            model = section.model_stiffness_matrix()
        self._section = section

        super().__init__(*args, model=model, **kwargs)

        if self.celldata is not None:
            nE = len(self.celldata)
            if areas is None:
                if isinstance(self.section, BeamSection):
                    areas = np.full(nE, self.section.A)
                else:
                    areas = np.ones(nE)
            else:
                assert len(areas.shape) == 1, \
                    "'areas' must be a 1d float or integer numpy array!"
            dbkey = self.celldata._dbkey_areas_
            self.celldata.db[dbkey] = areas

    def simplify(self, inplace=True) -> 'LineMesh':
        pass

    @property
    def section(self) -> BeamSection:
        """
        Returns the section of the cells or None if there is no associated data.

        Returns
        -------
        BeamSection
            The section instance associated with the beams of the block.

        """
        if self._section is not None:
            return self._section
        else:
            if self.is_root():
                return self._section
            else:
                return self.parent.section

    def plot(self, *args, scalars=None, backend='plotly', scalar_labels=None, **kwargs):
        """
        Plots the line elements using one of the supported backends.

        Parameters
        ----------
        scalars : numpy.ndarray, Optional
            Data to plot. Default is None.

        backend : str, Optional
            The backend to use for plotting. Available options are 'plotly' and 'vtk'.
            Default is 'plotly'.

        scalar_labels : Iterable, Optional
            Labels of the scalars in 'scalars'. Only if Plotly is selected as the backend.
            Defaeult is None.

        Returns
        -------
        Any
            A PyVista or a Plotly object.

        """
        if backend == 'vtk':
            return self.pvplot(*args, scalars=scalars, scalar_labels=scalar_labels,
                               **kwargs)
        elif backend == 'plotly':
            assert __hasplotly__
            coords = self.coords()
            topo = self.topology()
            return plot_lines_3d(coords, topo)
        elif backend == 'mpl':
            raise NotImplementedError
            assert __hasmatplotlib__
        else:
            msg = "No implementation for backend '{}'".format(backend)
            raise NotImplementedError(msg)

    def plot_dof_solution(self, *args, backend: str = 'plotly', case: int = 0, component: int = 0,
                          labels: Iterable = None, **kwargs) -> Any:
        """
        Plots degrees of freedom solution using 'vtk' or 'plotly'.

        Parameters
        ----------
        scalars : numpy.ndarray, Optional
            Data to plot. Default is None.

        case : int, Optional
            The index of the load case. Default is 0.

        component : int, Optional
            The index of the DOF component. Default is 0.

        labels : Iterable, Optional
            Labels of the DOFs. Only if Plotly is selected as the backend.
            Defaeult is None.

        **kwargs : dict, Optional
            Keyqord arguments forwarded to :func:`pvplot`.

        Returns
        -------
        Any
            A figure object or None, depending on the selected backend.

        """
        if backend == 'vtk':
            scalars = self.nodal_dof_solution()[:, component, case]
            return self.pvplot(*args, scalars=scalars, **kwargs)
        elif backend == 'plotly':
            scalar_labels = labels if labels is not None else ['U', 'V', 'W']
            coords = self.coords()
            topo = self.topology()
            dofsol = self.nodal_dof_solution()[:, :3, case]
            return plot_lines_3d(coords, topo, scalars=dofsol, scalar_labels=scalar_labels)
        else:
            msg = "No implementation for backend '{}'".format(backend)
            raise NotImplementedError(msg)

    def _init_config_(self):
        super()._init_config_()
        key = self.__class__._pv_config_key_
        self.config[key]['color'] = 'k'
        self.config[key]['line_width'] = 10
        self.config[key]['render_lines_as_tubes'] = True


class BernoulliFrame(LineMesh):
    """
    A subclass of :class:`LineMesh` to handle input and output
    of 1d meshes. 

    Note
    ----
    This in experimental stage.  

    """

    NDOFN = 6

    _cell_classes_ = {
        2: B2,
        3: B3,
    }



# -*- coding: utf-8 -*-
from typing import Union, Tuple
from numpy import ndarray
from scipy.sparse import spmatrix
import numpy as np

from neumann.array import atleast2d

from .linsolve import solve_standard_form
from .eigsolve import eig_sparse, eig_dense


ArrayLike = Union[ndarray, spmatrix]


def effective_modal_mass(M: ArrayLike, action: ndarray,
                         mode: ndarray) -> float:
    """
    Returns the effective modal mass for a specific mode.

    Assumes that the modal shapes are normalized to the mass matrix.

    Parameters
    ----------
    M : numpy.ndarray
        2d mass matrix as a NumPy or SciPy 2d float array.
    action : Iterable
        1d iterable, with a length matching the dof layout of the structure. 
    mode : numpy.ndarray
        1d array representing a modal shape.

    Returns
    -------
    float  
        The effective modal mass.

    """
    return (mode.T @ M @ action)**2


def effective_modal_masses(M: ArrayLike, action: ndarray,
                           modes: ndarray) -> ndarray:
    """
    Returns the effective modal mass for several modes.

    Assumes that the modal shapes are normalized to the mass matrix.

    Parameters
    ----------
    M : numpy.ndarray
        Mass matrix as a NumPy or SciPy 2d float array.
    action : numpy.ndarray
        1d iterable, with a length matching the dof layout of the structure. 
    modes : numpy.ndarray
        A matrix, whose columns are eigenmodes of interest.

    Notes
    -----
    The sum of all effective masses equals the total mass of the structure.

    Returns
    -------
    numpy.ndarray  
        1d float array of effective modal masses

    """
    res = []
    for i in range(modes.shape[-1]):
        m_eff = effective_modal_mass(M, action, modes[:, i])
        res.append(m_eff)
    return np.array(res)


def Rayleigh_quotient(M: spmatrix, *, K: spmatrix = None, u: ndarray = None,
                      f: ndarray = None, **kw) -> ndarray:
    """
    Returns Rayleigh's quotient

    .. math::
        :nowrap:

        \\begin{equation}
            \\frac{\\mathbf{v}^T \\mathbf{K} \\mathbf{v}}{\\mathbf{v}^T 
            \\mathbf{M} \\mathbf{v}},
        \\end{equation}

    for a prescribed action :math:`\\mathbf{v}`.

    Parameters
    ----------
    M : scipy.linalg.sparse.spmatrix
        The mass matrix.
    u : numpy.ndarray, Optional
        The vector of nodal displacements (1d or 2d).
    K : scipy.linalg.sparse.spmatrix, Optional
        The stiffness matrix. It can be omitted, if 'f' is provided.
    f : numpy.ndarray, Optional
        The vector of nodal forces (1d or 2d). It can be omitted, 
        if 'K' is provided.

    Notes
    -----
    The Rayleigh quotient is 

    - higher than, or equal to the square of the smallest natural 
      circular frequency,

    - smaller than, or equal to the square of the highest natural 
      circular frequency.

    Returns
    -------
    numpy.ndarray
        An 1d array of floats.

    """
    if isinstance(f, ndarray) and u is None:
        u = atleast2d(solve_standard_form(K, f), back=True)
    elif isinstance(u, ndarray) and f is None:
        f = K @ u
    res = []
    for i in range(u.shape[-1]):
        nom = u[:, i].T @ f[:, i]
        denom = u[:, i].T @ M @ u[:, i]
        res.append(nom / denom)
    return np.array(res)


def natural_circular_frequencies(K: spmatrix, M: spmatrix, *, k: int = 10,
                                 return_vectors: bool = False,  maxiter: int = 5000,
                                 normalize: bool = True, as_dense: bool = False,
                                 around: float = None, nmode: str = 'M', 
                                 which: str = 'SM', **kwargs) -> Tuple[ndarray]:
    """
    Returns the natural circular frequencies :math:`\omega_{0i}` and optionally the 
    corresponding eigenvectors as (not trivial) solutions to the eigenproblem

    .. math::
        :nowrap:

        \\begin{equation}
            \\left( \\mathbf{K} - \\omega^2 \\mathbf{M} \\right) 
            \\mathbf{v} = \\mathbf{0}. 
        \\end{equation}

    Parameters
    ----------
    K : scipy.linalg.sparse.spmatrix
        The stiffness matrix in sparse format. 
    M : scipy.linalg.sparse.spmatrix
        The mass matrix in sparse format.
    k : int, Optional
        The number of solutions to return. Only if 'as_dense' is False. 
        Default is 10.
    return_vectors : bool, Optional
        To return eigenvectors or not. Default is False.
    maxiter : int, Optional
        Maximum number of iterations, if solved using sparse matrices.
        Only if 'as_dense' is False. Default is 5000.
    normalize : bool, Optional
        If True, the returned eigenvectors are normalized to the mass matrix.
        Only if 'return_vectors' is True. Default is True.
    as_dense : bool, Optional
        If True, the stiffness and mass matrices are handled as dense arrays.
        In this case, if 'return_vectors' is True, all eigenvalues and vectors 
        are returned, regardless of other parameters.
    around : float, Optional
        A target (possibly an approximation) value around which eigenvalues
        are searched. Default is None.
    which : str, ['LM' | 'SM' | 'LA' | 'SA' | 'BE' | 'LR' | 'SR' 
                 | 'LI' | 'SI'], Optional
        Which `k` eigenvectors and eigenvalues to find:

            'LM' : largest magnitude

            'SM' : smallest magnitude

            'LA' : Largest (algebraic) eigenvalues.

            'SA' : Smallest (algebraic) eigenvalues.

            'BE' : Half (k/2) from each end of the spectrum.

            'LR' : largest real part

            'SR' : smallest real part

            'LI' : largest imaginary part

            'SI' : smallest imaginary part

        Only if 'as_dense' is False. Default is 'LM'.
    **kwargs : dict, Optional
        Keyword arguments are forwarded to the appropriate routine of SciPy.

    Note
    ----
    Calculation is done using the relavant routines of Scipy,
    see its documentation for more control over the parameters.

    See also
    --------
    :func:`scipy.linalg.eigs`
    :func:`scipy.sparse.linalg.eigsh`

    Returns
    -------
    numpy.ndarray
        The natural circular frequencies.
    numpy.ndarray
        The eigenvectors, only if 'return_vectors' is True.

    """
    if around is not None:
        kwargs['sigma'] = around**2

    if as_dense:
        vals, vecs = eig_dense(K, M=M, nmode=nmode, normalize=normalize,
                               return_residuals=False, **kwargs)
    else:
        vals, vecs = eig_sparse(K, k=k, M=M, which=which,
                                normalize=normalize, maxiter=maxiter,
                                return_residuals=False, nmode=nmode,
                                **kwargs)
    cfreqs = np.sqrt(vals)
    if return_vectors:
        return cfreqs, vecs
    return cfreqs


def estimate_smallest_natural_circular_frequency(*args, **kwargs) -> ndarray:
    """
    Returns a lower bound estimation of the smallest natural frequency using 
    Rayleigh's quotient. See :func:`Rayleigh_quotient` for the input.

    Notes
    -----
    This relies on the equivalence of elastic internal energy and kinetic 
    energy of undamped conservative systems.

    Returns
    -------
    numpy.ndarray
        An 1d array of floats.

    See also
    --------
    :func:`Rayleigh_quotient`

    """
    return np.sqrt(Rayleigh_quotient(*args, **kwargs))



# -*- coding: utf-8 -*-
from typing import Iterable, Union, Tuple

import numpy as np
from numba import njit, prange
from numpy import ndarray, sin, cos, ndarray, pi as PI

from neumann import squeeze
from neumann.array import atleast1d, atleast2d, atleast3d


@squeeze(True)
def lhs_Navier(size: Union[float, Tuple[float]], shape: Union[int, Tuple[int]], *,
               D: Union[float, ndarray], S: Union[float, ndarray] = None, **kw):
    """
    Returns coefficient matrices for a Navier solution, for a single or 
    multiple left-hand sides.

    Parameters
    ----------
    size : Union[float, Tuple[float]]
        The size of the problem. Scalar for a beam, 2-tuple for a plate.

    shape : Union[int, Tuple[int]]
        The number of harmonic terms used. Scalar for a beam, 2-tuple for a plate.

    D : Union[float, ndarray]
        2d or 3d float array of bending stiffnesses for a plate, scalar or 1d float array
        for a beam.

    S : Union[float, ndarray], Optional
        2d or 3d float array of shear stiffnesses for a plate, scalar or 1d float array
        for a beam. Only for Mindlin-Reissner plates and Euler-Bernoulli beams.
        plates. Default is None.

    squeeze : bool, Optional
        Removes single-dimensional entries from the shape of the 
        resulting array. Default is True.
        
    Note
    ----
    Shear stiffnesses must include shear correction.

    Returns
    -------
    numpy.ndarray
        The coefficients as an array. See the documentation of the corresponding
        function for further details. Also, if `squeeze` is True, axes of length one 
        are removed.

    See Also
    --------
    :func:`lhs_Navier_Mindlin`
    :func:`lhs_Navier_Kirchhoff`
    :func:`lhs_Navier_Bernoulli`
    :func:`lhs_Navier_Timoshenko` 

    """
    if isinstance(shape, Iterable):  # plate problem
        if S is None:
            return lhs_Navier_Kirchhoff(size, shape, atleast3d(D))
        else:
            return lhs_Navier_Mindlin(size, shape, atleast3d(D), atleast3d(S))
    else:  # beam problem
        if S is None:
            return lhs_Navier_Bernoulli(size, shape, atleast1d(D))
        else:
            return lhs_Navier_Timoshenko(size, shape, atleast1d(D), atleast1d(S))


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Mindlin(size: tuple, shape: tuple, D: np.ndarray, S: np.ndarray):
    """
    JIT compiled function, that returns coefficient matrices for a Navier 
    solution for multiple left-hand sides.

    Parameters
    ----------
    size : tuple
        Tuple of floats, containing the sizes of the rectagle.

    shape : tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.

    D : numpy.ndarray
        3d float array of bending stiffnesses.

    S : numpy.ndarray
        3d float array of shear stiffnesses.
        
    Note
    ----
    The shear stiffness values must include the shear correction factor.

    Returns
    -------
    numpy.ndarray
        4d float array of coefficients.

    """
    Lx, Ly = size
    nLHS = D.shape[0]
    M, N = shape
    res = np.zeros((nLHS, M * N, 3, 3), dtype=D.dtype)
    for iLHS in prange(nLHS):
        D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
            D[iLHS, 1, 1], D[iLHS, 2, 2]
        S44, S55 = S[iLHS, 0, 0], S[iLHS, 1, 1]
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                iMN = (m - 1) * N + n - 1
                res[iLHS, iMN, 0, 0] = -PI**2*D22*n**2 / \
                    Ly**2 - PI**2*D66*m**2/Lx**2 - S44
                res[iLHS, iMN, 0, 1] = PI**2*D12*m * \
                    n/(Lx*Ly) + PI**2*D66*m*n/(Lx*Ly)
                res[iLHS, iMN, 0, 2] = PI*S44*n/Ly
                res[iLHS, iMN, 1, 0] = -PI**2*D12*m * \
                    n/(Lx*Ly) - PI**2*D66*m*n/(Lx*Ly)
                res[iLHS, iMN, 1, 1] = PI**2*D11*m**2 / \
                    Lx**2 + PI**2*D66*n**2/Ly**2 + S55
                res[iLHS, iMN, 1, 2] = PI*S55*m/Lx
                res[iLHS, iMN, 2, 0] = -PI*S44*n/Ly
                res[iLHS, iMN, 2, 1] = PI*S55*m/Lx
                res[iLHS, iMN, 2, 2] = PI**2*S44 * \
                    n**2/Ly**2 + PI**2*S55*m**2/Lx**2
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Kirchhoff(size: tuple, shape: tuple, D: np.ndarray):
    """
    JIT compiled function, that returns coefficient matrices for a Navier 
    solution for multiple left-hand sides.

    Parameters
    ----------
    size : tuple
        Tuple of floats, containing the sizes of the rectagle.

    shape : tuple
        Tuple of integers, containing the number of harmonic terms
        included in both directions.

    D : numpy.ndarray
        3d float array of bending stiffnesses.

    Returns
    -------
    numpy.ndarray
        2d float array of coefficients.

    """
    Lx, Ly = size
    nLHS = D.shape[0]
    M, N = shape
    res = np.zeros((nLHS, M * N), dtype=D.dtype)
    for iLHS in prange(nLHS):
        D11, D12, D22, D66 = D[iLHS, 0, 0], D[iLHS, 0, 1], \
            D[iLHS, 1, 1], D[iLHS, 2, 2]
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                iMN = (m - 1) * N + n - 1
                res[iLHS, iMN] = PI**4*D11*m**4/Lx**4 + \
                    2*PI**4*D12*m**2*n**2/(Lx**2*Ly**2) + \
                    PI**4*D22*n**4/Ly**4 + \
                    4*PI**4*D66*m**2*n**2/(Lx**2*Ly**2)
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Bernoulli(L: float, N: int, EI: ndarray):
    """
    JIT compiled function, that returns coefficient matrices for a Navier 
    solution for multiple left-hand sides.

    Parameters
    ----------
    L : float
        The length of the beam.

    N : int
        The number of harmonic terms.

    EI : numpy.ndarray
        1d float array of bending stiffnesses.

    Returns
    -------
    numpy.ndarray
        2d float array of coefficients.

    """
    nLHS = EI.shape[0]
    res = np.zeros((nLHS, N), dtype=EI.dtype)
    for iLHS in prange(nLHS):
        for n in prange(1, N + 1):
            res[iLHS, n - 1] = PI**4*EI[iLHS]*n**4/L**4
    return res


@njit(nogil=True, parallel=True, cache=True)
def lhs_Navier_Timoshenko(L: float, N: int, EI: ndarray, GA: ndarray):
    """
    JIT compiled function, that returns coefficient matrices for a Navier 
    solution for multiple left-hand sides.

    Parameters
    ----------
    L : float
        The length of the beam.

    N : int
        The number of harmonic terms.

    EI : numpy.ndarray
        1d float array of bending stiffnesses.
        
    GA : numpy.ndarray
        1d float array of shear stiffnesses.

    Note
    ----
    The shear stiffness values must include the shear correction factor.
    
    Returns
    -------
    numpy.ndarray
        4d float array of coefficients.

    """
    nLHS = EI.shape[0]
    res = np.zeros((nLHS, N, 2, 2), dtype=EI.dtype)
    for iLHS in prange(nLHS):
        for n in prange(1, N + 1):
            iN = n - 1
            c1 = PI * n / L
            c2 = PI**2 * n**2 / L**2
            res[iLHS, iN, 0, 0] = c2 * GA[iLHS]
            res[iLHS, iN, 0, 1] = - c1 * GA[iLHS]
            res[iLHS, iN, 1, 0] = res[iLHS, iN, 0, 1]
            res[iLHS, iN, 1, 1] = GA[iLHS] + c2 * EI[iLHS]
    return res


@njit(nogil=True, parallel=True, cache=True)
def rhs_Bernoulli(coeffs: ndarray, L: float):
    """
    Calculates unknowns for Bernoulli Beams.
    """
    nRHS, N = coeffs.shape[:2]
    res = np.zeros((nRHS, N))
    c = PI / L
    for i in prange(nRHS):
        for n in prange(N):
            res[i, n] = coeffs[i, n, 0] + coeffs[i, n, 1] * c * (n + 1)
    return res


@squeeze(True)
def rhs_line_const(L: float, N: int, x: ndarray, values: ndarray):
    """
    Returns coefficients for constant loads over line segments in 
    the order [f, m].    
    """
    return _line_const_(L, N, atleast2d(x), atleast2d(values))


@njit(nogil=True, parallel=True, cache=True)
def _line_const_(L: float, N: int, x: ndarray, values: ndarray):
    nR = values.shape[0]
    rhs = np.zeros((nR, N, 2), dtype=x.dtype)
    for iR in prange(nR):
        for n in prange(1, N + 1):
            iN = n - 1
            xa, xb = x[iR]
            f, m = values[iR]
            c = PI * n / L
            rhs[iR, iN, 0] = (2 * f / (PI*n)) * (cos(c*xa) - cos(c*xb))
            rhs[iR, iN, 1] = (2 * m / (PI*n)) * (sin(c*xb) - sin(c*xa))
    return rhs


@squeeze(True)
def rhs_rect_const(size: tuple, shape: tuple, values: np.ndarray,
                   points: np.ndarray):
    """
    Returns coefficients for constant loads over rectangular patches 
    in the order [f, mx, my].    
    """
    return _rect_const_(size, shape, atleast2d(values), atleast3d(points))


@njit(nogil=True, cache=True)
def __rect_const__(size: tuple, m: int, n: int, xc: float, yc: float,
                   w: float, h: float, values: ndarray):
    Lx, Ly = size
    f, mx, my = values
    return np.array([
        16*f*sin((1/2)*PI*m*w/Lx)*sin(PI*m*xc/Lx) *
        sin((1/2)*PI*h*n/Ly)*sin(PI*n*yc/Ly)/(PI**2*m*n),
        16*mx*sin((1/2)*PI*m*w/Lx) *
        sin((1/2)*PI*h*n/Ly)*sin(PI*n*yc/Ly) *
        cos(PI*m*xc/Lx)/(PI**2*m*n),
        16*my*sin((1/2)*PI*m*w/Lx) *
        sin(PI*m*xc/Lx)*sin((1/2)*PI*h*n/Ly) *
        cos(PI*n*yc/Ly)/(PI**2*m*n)
    ])


@njit(nogil=True, parallel=True, cache=True)
def _rect_const_(size: tuple, shape: tuple, values: ndarray,
                 points: ndarray):
    nR = values.shape[0]
    M, N = shape
    rhs = np.zeros((nR, M * N, 3), dtype=points.dtype)
    for iR in prange(nR):
        xmin, ymin = points[iR, 0]
        xmax, ymax = points[iR, 1]
        xc = (xmin + xmax)/2
        yc = (ymin + ymax)/2
        w = np.abs(xmax - xmin)
        h = np.abs(ymax - ymin)
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                mn = (m - 1) * N + n - 1
                rhs[iR, mn, :] = __rect_const__(size, m, n, xc, yc,
                                                w, h, values[iR])
    return rhs


@squeeze(True)
def rhs_conc_1d(L: float, N: int, values: ndarray, points: ndarray):
    return _conc1d_(L, N, atleast2d(values), atleast1d(points))


@njit(nogil=True, parallel=True, cache=True)
def _conc1d_(L: tuple, N: tuple, values: ndarray, points: ndarray):
    nRHS = values.shape[0]
    c = 2 / L
    rhs = np.zeros((nRHS, N, 2), dtype=points.dtype)
    PI = np.pi
    for iRHS in prange(nRHS):
        x = points[iRHS]
        f, m = values[iRHS]
        Sx = PI * x / L
        for n in prange(1, N + 1):
            i = n - 1
            rhs[iRHS, i, 0] = c * f * sin(n * Sx)
            rhs[iRHS, i, 1] = c * m * cos(n * Sx)
    return rhs


@squeeze(True)
def rhs_conc_2d(size: tuple, shape: tuple, values: ndarray, points: ndarray):
    return _conc2d_(size, shape, atleast2d(values), atleast2d(points))


@njit(nogil=True, parallel=True, cache=True)
def _conc2d_(size: tuple, shape: tuple, values: ndarray, points: ndarray):
    nRHS = values.shape[0]
    Lx, Ly = size
    c = 4 / Lx / Ly
    M, N = shape
    rhs = np.zeros((nRHS, M * N, 3), dtype=points.dtype)
    PI = np.pi
    for iRHS in prange(nRHS):
        x, y = points[iRHS]
        fz, mx, my = values[iRHS]
        Sx = PI * x / Lx
        Sy = PI * y / Ly
        for m in prange(1, M + 1):
            for n in prange(1, N + 1):
                mn = (m - 1) * N + n - 1
                rhs[iRHS, mn, 0] = c * fz * sin(m * Sx) * sin(n * Sy)
                rhs[iRHS, mn, 1] = c * mx * cos(m * Sx) * sin(n * Sy)
                rhs[iRHS, mn, 2] = c * my * sin(m * Sx) * cos(n * Sy)
    return rhs



# -*- coding: utf-8 -*-
import numpy as np
from linkeddeepdict import LinkedDeepDict
from linkeddeepdict.tools import getallfromkwargs

from polymesh.space import StandardFrame, PointCloud
from polymesh.grid import gridH8 as grid
from polymesh.topo.tr import H8_to_L2
from polymesh.space import frames_of_lines
from polymesh.space.utils import index_of_closest_point

from sigmaepsilon.solid import BeamSection
from sigmaepsilon.solid import Structure, LineMesh, PointData
from sigmaepsilon.solid.fem.cells import B2 as Beam


def console_grid_bernoulli() -> Structure:
    """
    Returns a simple console as a grid of 2-noded Bernoulli beams.

    Example
    -------
    >>> from sigmaepsilon.examples import console_grid_bernoulli
    >>> structure = console_grid_bernoulli()
    >>> structure.linsolve()

    """
    # units in kN and cm

    Lx, Ly, Lz = 100., 15., 10.  # size of the grid
    nx, ny, nz = 4, 2, 2  # desity of the grid

    d = 1.2  # outer diameter of the tube
    t = 0.4  # thickness of the tube

    Ex = 21000.0
    nu = 0.25
    fy = 35.0

    loads = LinkedDeepDict()

    # load case 1
    loads['LC1', 'position'] = [Lx, 0, 0]
    loads['LC1', 'value'] = [0, 0, 2.0, 0, 0, 0]

    # load case 2
    loads['LC2', 'position'] = [Lx, Ly/2, 0]
    loads['LC2', 'value'] = [0, 2.0, 0, 0, 0, 0]

    # load case 3
    loads['LC3', 'position'] = [Lx, -Ly/2, 0]
    loads['LC3', 'value'] = [0, -2.0, 0, 0, 0, 0]

    gridparams = {
        'size': (Lx, Ly, Lz),
        'shape': (nx, ny, nz),
        'origo': (0, 0, 0),
        'start': 0
    }

    coords, topo = grid(**gridparams)
    coords, topo = H8_to_L2(coords, topo)

    GlobalFrame = StandardFrame(dim=3)

    points = PointCloud(coords, frame=GlobalFrame).centralize()
    dx = - np.array([points[:, 0].min(), 0., 0.])
    points.move(dx)
    coords = points.show()

    section = BeamSection('CHS', d=d, t=t, n=16)
    section.calculate_section_properties()
    section_props = section.section_properties
    A, Ix, Iy, Iz = getallfromkwargs(['A', 'Ix', 'Iy', 'Iz'],
                                     **section_props)

    G = Ex / (2 * (1 + nu))
    Hooke = np.array([
        [Ex*A, 0, 0, 0],
        [0, G*Ix, 0, 0],
        [0, 0, Ex*Iy, 0],
        [0, 0, 0, Ex*Iz]
    ])

    # essential boundary conditions
    ebcinds = np.where(coords[:, 0] < 1e-12)[0]
    fixity = np.zeros((coords.shape[0], 6)).astype(bool)
    fixity[ebcinds, :] = True

    # natural boundary conditions
    nLoadCase = len(loads)
    nodal_loads = np.zeros((coords.shape[0], 6, nLoadCase))
    for iLC, key in enumerate(loads):
        x = loads[key]['position']
        f = loads[key]['value']
        iN = index_of_closest_point(coords, np.array(x))
        loads[key]['node'] = iN
        nodal_loads[iN, :, iLC] = f

    # pointdata
    pd = PointData(coords=coords, frame=GlobalFrame,
                   loads=nodal_loads, fixity=fixity)

    # celldata
    frames = frames_of_lines(coords, topo)
    cd = Beam(topo=topo, frames=frames)

    # set up mesh and structure
    mesh = LineMesh(pd, cd, model=Hooke, frame=GlobalFrame)
    structure = Structure(mesh=mesh)

    return structure



# -*- coding: utf-8 -*-
import numpy as np

from polymesh.cells import T6 as Triangle

from ..model.membrane import Membrane
from ..model.plate import Plate

from .elem import FiniteElement
from .metaelem import ABCFiniteElement as ABC


class LSTM(ABC, Membrane, Triangle, FiniteElement):

    qrule = 'selective'
    quadrature = {
        'full': (np.array([[1/6, 1/6], [2/3, 1/6], [1/6, 2/3]]),
                 np.array([1/6, 1/6, 1/6])),
        'full2': (np.array([[1/2, 1/2], [0, 1/2], [1/2, 0]]),
                  np.array([1/6, 1/6, 1/6])),
        'selective': {
            (0, 1): 'full',
            (2,): 'reduced'
        },
        'reduced': (np.array([[1/3, 1/3]]), np.array([1/2]))
    }


class LSTP(ABC, Plate, Triangle, FiniteElement):

    qrule = 'selective'
    quadrature = {
        'full': (np.array([[1/6, 1/6], [2/3, 1/6], [1/6, 2/3]]),
                 np.array([1/6, 1/6, 1/6])),
        'full2': (np.array([[1/2, 1/2], [0, 1/2], [1/2, 0]]),
                  np.array([1/6, 1/6, 1/6])),
        'selective': {
            (0, 1, 2): 'full',
            (3, 4): 'reduced'
        },
        'reduced': (np.array([[1/3, 1/3]]), np.array([1/2]))
    }



# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numpy import ndarray
from typing import Union, Iterable

from .solid import Solid

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix(gdshp: ndarray):
    """
    Returns the matrix expressing the relationship of generalized strains
    and generalized displacements of a Bernoulli beam element. The order
    of strain components is
        0 : strain along x
        1 : curvature around x
        2 : curvature around y
        3 : curvature around z 

    Parameters
    ----------
    gdshp : numpy.ndarray  
        First, second and third derivatives for every dof(6) of every node(nNE)
        as a 3d float array of shape (nNE, nDOF=6, 3).

    Returns
    -------
    numpy.ndarray  
        Apprpximation coefficients for every generalized strain and every 
        shape function as a 2d float array of shape (nSTRE=4, nNE * 6).
    
    """
    nNE = gdshp.shape[0]
    B = np.zeros((4, nNE * 6), dtype=gdshp.dtype)
    for i in prange(nNE):
        di = i * 6
        # \epsilon_x
        B[0, 0 + di] = gdshp[i, 0, 0]
        # \kappa_x
        B[1, 3 + di] = gdshp[i, 3, 0]
        # \kappa_y
        B[2, 2 + di] = -gdshp[i, 2, 1]
        B[2, 4 + di] = -gdshp[i, 4, 1]
        # \kappa_z
        B[3, 1 + di] = gdshp[i, 1, 1]
        B[3, 5 + di] = gdshp[i, 5, 1]
    return B


@njit(nogil=True, parallel=True, cache=__cache)
def strain_displacement_matrix_bulk(gdshp: ndarray):
    """
    Calculates the strain-displacement matrix for several elements.
    
    Parameters
    ----------
    gdshp : numpy.ndarray
        5d float array of shape (nE, nP, nNE, nDOF=6, 3).

    Returns
    -------
    numpy.ndarray
        4d float array of shape (nE, nP, nSTRE=4, nNODE * nDOF=6).

    """
    nE, nP, nNE = gdshp.shape[:3]
    B = np.zeros((nE, nP, 4, nNE * 6), dtype=gdshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            B[iE, iP] = strain_displacement_matrix(gdshp[iE, iP])
    return B


@njit(nogil=True, parallel=True, cache=__cache)
def calculate_shear_forces(dofsol: ndarray, forces: ndarray,
                           D: ndarray, gdshp: ndarray):
    # dofsol (nE, nNE, nDOF=6, nRHS)
    # forces (nE, nP, 4, nRHS)
    # gdshp (nE, nP, nNE=2, nDOF=6, 3)
    nE, nP, _, nRHS = forces.shape
    nNE = dofsol.shape[1]
    res = np.zeros((nE, nP, 6, nRHS), dtype=forces.dtype)
    res[:, :, 0, :] = forces[:, :, 0, :]
    res[:, :, 3, :] = forces[:, :, 1, :]
    res[:, :, 4, :] = forces[:, :, 2, :]
    res[:, :, 5, :] = forces[:, :, 3, :]
    for i in prange(nE):
        for j in prange(nP):
            for k in prange(nRHS):
                for m in range(nNE):
                    # Vy
                    res[i, j, 1, k] -= D[i, 3, 3] * (
                        gdshp[i, j, m, 1, 2] * dofsol[i, m, 1, k] +
                        gdshp[i, j, m, 5, 2] * dofsol[i, m, 5, k])
                    # Vz
                    res[i, j, 2, k] -= D[i, 2, 2] * (
                        gdshp[i, j, m, 2, 2] * dofsol[i, m, 2, k] +
                        gdshp[i, j, m, 4, 2] * dofsol[i, m, 4, k])
    return res


class BernoulliBeam(Solid):

    dofs = ('UX', 'UY', 'UZ', 'ROTX', 'ROTY', 'ROTZ')

    NDOFN = 6
    NSTRE = 4

    def model_stiffness_matrix(self, *args, **kwargs) -> ndarray:
        """
        Returns the model stiffness matrix for all elements in the block.
                
        Returns
        -------
        numpy.ndarray
            3d float array of shape (nE, nSTRE, nSTRE), where nE and nSTRE
            are the number of cells and stress/strain components.
            
        """
        return self.material_stiffness_matrix()

    def strain_displacement_matrix(self, pcoords:Union[float, Iterable[float]]=None, *args, 
                                   rng:Iterable=None, jac: ndarray = None, dshp: ndarray = None,
                                   **kwargs) -> ndarray:
        """
        Calculates the strain displacement matrix.
        
        Parameters
        ----------
        pcoords : float or Iterable, Optional
            Locations of the evaluation points. Default is None.

        rng : Iterable, Optional
            The range in which the locations ought to be understood, 
            typically [0, 1] or [-1, 1]. Only if 'pcoords' is provided.
            Default is [0, 1].

        jac : Iterable, Optional
            The jacobian matrix as a float array of shape (nE, nP, 1, 1), evaluated for 
            each point in each cell. Default is None.

        dshp : Iterable, Optional
            The shape function derivatives of the master element as a float array of 
            shape (nP, nNE, nDOF=6, 3), evaluated at a 'nP' number of points. 
            Default is None.
            
        """
        rng = np.array([-1, 1]) if rng is None else np.array(rng)
        gdshp = self.shape_function_derivatives(
            pcoords, rng=rng, jac=jac, dshp=dshp)
        return strain_displacement_matrix_bulk(gdshp)

    def masses(self, *args, values=None, **kwargs) -> ndarray:
        """
        Returns the masses of the cells in the block.
        """
        if isinstance(values, np.ndarray):
            dens = values
        else:
            dens = self.db.density
        try:
            areas = self.areas()
        except Exception:
            areas = np.ones_like(dens)
        lengths = self.lengths()
        return areas * dens * lengths

    def mass(self, *args, **kwargs) -> float:
        """
        Returns the total mass of the block.
        """
        return np.sum(self.masses(*args, **kwargs))



# -*- coding: utf-8 -*-
from typing import Iterable, Union, Tuple
from abc import abstractmethod
import numpy as np
from numpy import ndarray, concatenate as conc
from scipy.sparse import coo_matrix as coo

from neumann.linalg import ReferenceFrame
from neumann.array import atleast1d, atleast2d, repeat_diagonal_2d
from polymesh.space import index_of_closest_point

from .mesh import FemMesh
from .dofmap import DOF
from .constants import DEFAULT_DIRICHLET_PENALTY


__all__ = ['NodalSupport']


class EssentialBoundaryCondition:
    """
    Base class for Dirichlet boundary conditions accounted for
    using Courant-type penalization. 
    """
    
    @abstractmethod
    def assemble(self, mesh:FemMesh) -> Tuple[coo, ndarray]:
        ...


class NodalSupport(EssentialBoundaryCondition):
    """
    A class to handle nodal supports.
    
    Parameters
    ----------
    x : Iterable, Optional
        An iterable expressing the location of one or more points. 
        Use it if you don't know the index of the point of application 
        in advance. Default is None.
        
    i : Iterable[int] or int, Optional
        The index of one or more points of in a pointcloud. 
        Default is None.
            
    values : dict, Optional
        Dictionary to define the prescribed values. Valid keys are
        'UX', 'UY', 'UZ', 'ROTX', 'ROTY', 'ROTZ'. Default is None.
    
    frame : numpy.ndarray or ReferenceFrame, Optional
        Use it do define inclined supports. If it is a NumPy array, 
        it must be a 3x3 matrix representing a set of orthonormal base 
        vectors. A None value means that the prescribed values are understood
        in the frame of the mesh the constraints are applied to. 
        Default is None.
        
    penalty : float, Optional
        Penalty value for Courant-type penalization. 
        Default is `sigmaepsilon.solid.fem.constants.DEFAULT_DIRICHLET_PENALTY`.
        
    **kwargs : dict, Optional
        Keyword arguments used for elementwise definition of 'values'.
        
    Examples
    --------
    To define a support at one node
    
    >>> from sigmaepsilon.solid.fem import NodalSupport
    >>> support = NodalSupport(x=[10, 0, 0], UZ=1.)
    
    To define an inclined support with nonzero prescribed displacements:
    
    >>> from polymesh.space import StandardFrame
    >>> GlobalFrame = StandardFrame(dim=3)
    >>> frame = GlobalFrame.rotate_new('Body', [0, angle, 0], 'XYZ')
    >>> support = NodalSupport(x=[10, 0, 0], frame=frame, UX=1., UZ=1.)
    
    To apply the same constrain at multiple locations:
    
    >>> support = NodalSupport(x=[[0, 0, 0], [L, 0, 0]], UX = 0., UY=0.)
    
    To define a support with a custom penalty:
    
    >>> support = NodalSupport(x=[[0, 0, 0], [L, 0, 0]], 
    >>>                        UX = 0., UY=0., penalty=1e12)
            
    """

    def __init__(self, data:dict=None, *, x: Iterable = None, i: int = None,
                 frame: Union[ndarray, ReferenceFrame] = None, 
                 penalty : float = DEFAULT_DIRICHLET_PENALTY, **kwargs):
        self.x = x if x is None else np.array(x, dtype=float)
        self.i = i if i is None else np.array(i, dtype=int)
        self.penalty = penalty
        self.data = data
        self.frame = frame
        if data is None:
            self.data = kwargs
        self.dofmap = DOF.dofmap(self.data.keys())
        if not isinstance(self.i, int):
            assert self.x is not None, "Index or position must be defined!"
        if self.frame is not None:
            if not isinstance(self.frame, ReferenceFrame):
                raise TypeError("Invalid frame type. Read the " + \
                    "docs of the NodalSupport class.")
        else:
            self.frame = ReferenceFrame(dim=3)
            
    def assemble(self, mesh:FemMesh) -> Tuple[coo, ndarray]:
        """
        Returns the penalty stiffness matrix and the penalty load matrix.
        """
        i = None
        if self.i is None:
            coords = mesh.coords()
            x = atleast2d(self.x, front=True)
            i = index_of_closest_point(coords, x)
        else:
            i = atleast1d(self.i)
        nI = len(i)
        nDOF = mesh.NDOFN
        nN = len(mesh.pointdata)
        N = nDOF * nN
        c, r = divmod(nDOF, 3)
        assert r == 0, "The number of deegrees of freedom per" \
            + " node must be a multiple of 3."
        dcm = self.frame.dcm(source=mesh.frame)
        nodal_dcm = repeat_diagonal_2d(dcm, c)[self.dofmap]
        factors = repeat_diagonal_2d(nodal_dcm, nI)
        dofs = np.arange(nDOF)
        inds = conc([dofs + (i_ * nDOF) for i_ in i])
        kdata = (factors.T @ factors).flatten()
        krows = np.repeat(inds, len(inds))
        kcols = np.tile(inds, len(inds))
        Kp = self.penalty * coo((kdata, (krows, kcols)), shape=(N, N))
        fp = np.zeros(N, dtype=float)
        fdata = factors.T @ np.tile(list(self.data.values()), nI)
        fp[inds] = self.penalty * fdata
        Kp.eliminate_zeros()
        return Kp, fp



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

from neumann import squeeze
from neumann.linalg.solve import npsolve
from neumann.linalg import inv2x2, inv3x3


@squeeze(True)
def linsolve(LHS: ndarray, RHS: ndarray, *args, **kwarg):
    if len(LHS.shape) > 2:
        return linsolve2d(LHS, RHS)
    else:
        pass


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Bernoulli(A: ndarray, B: ndarray):
    """
    Calculates unknowns for Bernoulli Beams.
    """
    nLHS, N = A.shape
    nRHS = B.shape[0]
    res = np.zeros((nRHS, nLHS, N))
    for i in prange(nRHS):
        for j in prange(nLHS):
            for n in prange(N):
                res[i, j, n] = B[i, n] / A[j, n]
    return res


@njit(nogil=True, parallel=True, cache=True)
def linsolve_Timoshenko(A: ndarray, B: ndarray):
    """
    Calculates unknowns for Bernoulli Beams.
    """
    nLHS, nMN = A.shape[:2]
    nRHS = B.shape[0]
    res = np.zeros((nRHS, nLHS, nMN, 2))    
    for i in prange(nRHS):
        for j in prange(nLHS):
            for k in prange(nMN):
                res[i, j, k] = inv2x2(A[j, k]) @ B[i, k]
    return res


@njit(nogil=True, parallel=True, cache=True)
def linsolve2d(A: ndarray, B: ndarray):
    nLHS, nMN = A.shape[:2]
    nV = A.shape[-1]
    nRHS = B.shape[0]
    res = np.zeros((nRHS, nLHS, nMN, nV))
    if nV == 3:
        for i in prange(nRHS):
            for j in prange(nLHS):
                for k in prange(nMN):
                    res[i, j, k] = npsolve(A[j, k], B[i, k])
                    #res[i, j, k] = inv3x3(A[j, k]) @ B[i, k] 
    else:
        for i in prange(nRHS):
            for j in prange(nLHS):
                for k in prange(nMN):
                    res[i, j, k] = npsolve(A[j, k], B[i, k])
                    #res[i, j, k] = inv2x2(A[j, k]) @ B[i, k]
    return res


# -*- coding: utf-8 -*-
from numba import njit, prange
import numpy as np
from numba.typed import Dict as nbDict

from neumann.linalg.sparse.csr import csr_matrix as csr

__cache = True


@njit(nogil=True, parallel=True, cache=__cache)
def sensitivity_filter(sens: np.ndarray, dens: np.ndarray,
                       neighbours: nbDict, factors: dict):
    nE = len(dens)
    res = np.zeros_like(sens)
    inds = np.arange(nE).astype(np.int64)
    for iE in prange(nE):
        adj = neighbours[inds[iE]] 
        res[inds[iE]] = np.sum(factors[inds[iE]] * dens[adj] * sens[adj]) / \
            (dens[inds[iE]] * np.sum(factors[inds[iE]]))
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def sensitivity_filter_csr(sens: np.ndarray, dens: np.ndarray,
                           neighbours: csr, factors: dict):
    nE = len(dens)
    res = np.zeros_like(sens)
    for iE in prange(nE):
        adj, _ = neighbours.row(iE)
        res[iE] = np.sum(factors[iE] * dens[adj] * sens[adj]) / \
            (dens[iE] * np.sum(factors[iE]))
    return res



# -*- coding: utf-8 -*-
from typing import Union
from numba import njit, prange
import numpy as np
from numpy import ndarray
from numba import types as nbtypes
from numba.typed import Dict as nbDict

from neumann.linalg.sparse.csr import csr_matrix
from sigmaepsilon.solid.fem.utils import irows_icols_bulk_filtered

__cache = True

nbint64 = nbtypes.int64
nbint64A = nbint64[:]
nbfloat64A = nbtypes.float64[:]


def filter_stiffness(K_bulk: np.ndarray, edofs: np.ndarray,
                     vals: np.ndarray, *args, tol=1e-12, **kwargs):
    inds = np.where(vals > tol)[0]
    rows, cols = irows_icols_bulk_filtered(edofs, inds)
    return K_bulk[inds, :, :].flatten(), (rows.flatten(), cols.flatten())


@njit(nogil=True, cache=__cache)
def get_filter_factors(centers: np.ndarray, neighbours: Union[nbDict, ndarray], 
                       r_max: float):
    nE = len(centers)
    res = nbDict.empty(
        key_type=nbint64,
        value_type=nbfloat64A,
    )
    for iE in range(nE):
        res[iE] = r_max - norms(centers[neighbours[iE]] - centers[iE])
    return res


@njit(nogil=True, cache=__cache)
def get_filter_factors_csr(centers: np.ndarray, neighbours: csr_matrix,
                           r_max: float):
    nE = len(centers)
    res = nbDict.empty(
        key_type=nbint64,
        value_type=nbfloat64A,
    )
    for iE in range(nE):
        ni, _ = neighbours.row(iE)
        res[iE] = r_max - norms(centers[ni] - centers[iE])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def norms(a: np.ndarray):
    nI = len(a)
    res = np.zeros(nI, dtype=a.dtype)
    for iI in prange(len(a)):
        res[iI] = np.dot(a[iI], a[iI])
    return np.sqrt(res)


@njit(nogil=True, parallel=True, cache=__cache)
def norm(a: np.ndarray):
    return np.sqrt(np.dot(a, a))


@njit(nogil=True, parallel=True, cache=__cache)
def weighted_stiffness_bulk(K: np.ndarray, weights: np.ndarray):
    nE = len(weights)
    res = np.zeros_like(K, dtype=K.dtype)
    for i in prange(nE):
        res[i, :] = weights[i] * K[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def weighted_stiffness_1d(data: np.ndarray, weights: np.ndarray):
    nE = len(weights)
    dsize = int(len(data)/nE)
    res = np.zeros_like(data, dtype=data.dtype)
    for i in prange(nE):
        res[i*dsize: (i+1)*dsize] = data[i*dsize: (i+1)*dsize] * weights[i]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def compliances_bulk(K: np.ndarray, U: np.ndarray, gnum: np.ndarray):
    nE, nNE = gnum.shape
    res = np.zeros(nE, dtype=K.dtype)
    for iE in prange(nE):
        for i in prange(nNE):
            for j in prange(nNE):
                res[iE] += K[iE, i, j] * U[gnum[iE, i]] * U[gnum[iE, j]]
    return res



# -*- coding: utf-8 -*-
import numpy as np
from numpy import ndarray
from numba import njit, prange

__all__ = []

__cache__ = True
__NDOFN__ = 6


@njit(nogil=True, parallel=True, cache=__cache__)
def global_shape_function_derivatives_bulk(dshp: ndarray, jac: ndarray):
    """
    Calculates derivatives of shape functions w.r.t. the global axes
    using derivatives along local axes evaulated at some points in
    the interval [-1, 1], and jacobians of local-to-global mappings.

    Parameters
    ----------
    dshp: numpy.ndarray 
        A float array of shape (nE, nP, nNE, nDOF=6, 3)
        Derivatives of shape functions evaluated at an 'nP' number of 
        points of an 'nE' number of elements.

    jac: numpy.ndarray 
        A float array of shape (nE, nP, 1, 1)
        Jacobi determinants, evaluated for each point in each cell.

    Returns
    -------
    numpy.ndarray
        5d float array of shape (nE, nP, nNE, nDOF=6, 3).

    """
    nE, nP, nNE = dshp.shape[:3]
    res = np.zeros((nE, nP, nNE, __NDOFN__, 3), dtype=dshp.dtype)
    for iE in prange(nE):
        for jP in prange(nP):
            J = jac[iE, jP, 0, 0]
            res[iE, jP, :, :, 0] = dshp[iE, jP, :, :, 0] / J
            res[iE, jP, :, :, 1] = dshp[iE, jP, :, :, 1] / J**2
            res[iE, jP, :, :, 2] = dshp[iE, jP, :, :, 2] / J**3
    return res


@njit(nogil=True, parallel=True, cache=__cache__)
def shape_function_matrix(shp: ndarray, gdshp: ndarray):
    """
    Returns the shape function matrix of a single Bernoulli beam.

    The input contains evaluations of the shape functions and at least 
    the first global derivatives of the shape functions. 

    Parameters
    ----------
    shp: numpy.ndarray
        A float array of shape (nNE, nDOF=6), being the shape 
        functions evaluated at a single point of a single cell.

    gdshp: numpy.ndarray
        A float array of shape (nNE, nDOF=6, 3), being derivatives of shape 
        functions evaluated at a single point of a single cell.

    Returns
    -------
    numpy.ndarray
        2d float array of shape (nDOF, nDOF * nNODE).

    Notes
    -----
    The approximation applies a mixture of Lagrange(L) and Hermite(H) 
    polynomials, in the order [L, H, H, L, H, H].

    See Also
    --------
    :func:`shape_function_matrix_bulk`

    """
    nNE = shp.shape[0]
    res = np.zeros((__NDOFN__, nNE * __NDOFN__), dtype=gdshp.dtype)
    for iN in prange(nNE):
        di = iN * __NDOFN__
        # u_s
        res[0, 0 + di] = shp[iN, 0]
        # v_s
        res[1, 1 + di] = shp[iN, 1]
        res[1, 5 + di] = shp[iN, 5]
        # w_s
        res[2, 2 + di] = shp[iN, 2]
        res[2, 4 + di] = shp[iN, 4]
        # \heta_x
        res[3, 3 + di] = shp[iN, 3]
        # \theta_y
        res[4, 2 + di] = - gdshp[iN, 2, 0]
        res[4, 4 + di] = - gdshp[iN, 4, 0]
        # \theta_z
        res[5, 1 + di] = gdshp[iN, 1, 0]
        res[5, 5 + di] = gdshp[iN, 5, 0]
    return res


@njit(nogil=True, parallel=True, cache=__cache__)
def shape_function_matrix_bulk(shp: ndarray, gdshp: ndarray):
    """
    Returns the shape function matrix for several Bernoulli beams.

    The input contains evaluations of the shape functions and at least 
    the first global derivatives of the shape functions. 

    Parameters
    ----------
    shp: numpy.ndarray
        A float array of shape (nE, nP, nNE, nDOF=6), being the shape 
        functions evaluated at a nP number of points of several cells.

    dshp: numpy.ndarray
        A float array of shape (nE, nP, nNE, nDOF=6, 3), being derivatives 
        of shape functions evaluated at a nP number of points of several cells.

    Returns
    -------
    numpy.ndarray
        2d float array of shape (nE, nP, nDOF, nDOF * nNODE).

    See Also
    --------
    :func:`shape_function_matrix`

    """
    nE, nP, nNE = gdshp.shape[:3]
    res = np.zeros((nE, nP, __NDOFN__, nNE * __NDOFN__), dtype=gdshp.dtype)
    for iE in prange(nE):
        for iP in prange(nP):
            res[iE, iP] = shape_function_matrix(shp[iE, iP], gdshp[iE, iP])
    return res


@njit(nogil=True, parallel=True, cache=__cache__)
def _shape_function_matrix_L(shp: ndarray):
    """
    Returns the shape function matrix for a line.
    The input contains evaluations of the shape functions. 

    shp (nNE, nDOF=6)
    ---
    (nDOF, nDOF * nNODE)
    """
    nNE = shp.shape[0]
    res = np.zeros((__NDOFN__, nNE * __NDOFN__), dtype=shp.dtype)
    for iN in prange(nNE):
        di = iN * __NDOFN__
        res[0, 0 + di] = shp[iN, 0]
        res[1, 1 + di] = shp[iN, 1]
        res[2, 2 + di] = shp[iN, 2]
        res[3, 3 + di] = shp[iN, 3]
        res[4, 2 + di] = shp[iN, 4]
        res[5, 1 + di] = shp[iN, 5]
    return res


@njit(nogil=True, parallel=True, cache=__cache__)
def _shape_function_matrix_L_multi(shp: ndarray):
    """
    Returns the shape function matrix for a line.
    The input contains evaluations of the shape functions. 

    shp (nP, nNE, nDOF=6)
    ---
    (nP, nDOF, nDOF * nNODE)
    """
    nP, nNE = shp.shape[:2]
    res = np.zeros((nP, __NDOFN__, nNE * __NDOFN__), dtype=shp.dtype)
    for iP in prange(nP):
        res[iP] = _shape_function_matrix_L(shp[iP])
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache__)
def body_load_vector_bulk(values: ndarray, shp: ndarray, gdshp: ndarray,
                          djac: ndarray, w: ndarray):
    """
    Input values are assumed to be evaluated at multiple (nG) Gauss points of
    multiple (nE) cells.
    
    Parameters
    ----------
    values: numpy.ndarray
        A 3d float array of body loads of shape (nE, nRHS, nNE * 6)
        for several elements and load cases.
    
    shp: numpy.ndarray
        A float array of shape (nE, nG, nNE, nDOF=6), being the shape 
        functions evaluated at a nG number of Gauss points of several cells.

    djac: numpy.ndarray
        A float array of shape (nE, nG), being jacobian determinants
        evaluated at the Gauss points of several cells.
        
    w: numpy.ndarray
        1d float array of weights for an nG number of Gauss points,
        with a shape of (nG,).

    Returns
    -------
    numpy.ndarray
        3d float array of shape (nE, nNE * 6, nRHS).

    """
    nRHS = values.shape[1]
    nE, nG, nNE = shp.shape[:3]
    NH = shape_function_matrix_bulk(shp, gdshp)  # (nE, nG, nDOF, nDOF * nNODE)
    NL = _shape_function_matrix_L_multi(shp[0])  # (nG, nDOF, nDOF * nNODE)
    res = np.zeros((nE, nNE * __NDOFN__, nRHS), dtype=values.dtype)
    for iG in range(nG):
        for iRHS in prange(nRHS):
            for iE in prange(nE):
                res[iE, :, iRHS] += NH[iE, iG].T @ NL[iG] @ values[iE, iRHS, :] * \
                    djac[iE, iG] * w[iG]
    return res


@njit(nogil=True, parallel=True, cache=__cache__)
def lumped_mass_matrices_direct(dens: ndarray, lengths: ndarray, areas: ndarray,
                                topo: ndarray, alpha: float = 1/20) -> ndarray:
    """
    Returns the diagonal values of the directly lumped mass matrix for several 
    Bernoulli beam elements, sharing the total masses equivalently among nodes.

    Implementation is based on the lecture notes of Carlos A. Felippa [1].

    Parameters
    ----------
    dens : numpy.ndarray
        1d numpy float array of material densities.

    lengths : numpy.ndarray
        1d numpy float array of element lengths.

    areas : numpy.ndarray
        1d numpy float array of element areas.

    topo : numpy.ndarray
        2d numpy integer array of node indices referencing the global pointcloud.

    alpha : float, Optional
        A nonnegative parameter, typically between 0 and 1/50 (see notes).
        Default is 1/20.

    Returns
    -------
    numpy.ndarray
        A 2d float array of shape (nE, 6 * nNE), where nE and nNE are the number of
        elements and nodes per element. 

    Notes
    -----
    [1] : "The choice of α has been argued in the FEM literature over several decades, 
    but the whole discussion is largely futile. Matching the angular momentum of 
    the beam element gyrating about its midpoint gives α = −1/24. This violates the 
    positivity condition stated in [1, 32.2.4]. It follows that the best possible 
    α — as opposed to possible best — is zero. This choice gives, however, a singular 
    mass matrix, which is undesirable in scenarios where a mass-inverse appears."

    References
    ----------
    .. [1] Introduction to Finite Element Methods, Carlos A. Felippa.
           Department of Aerospace Engineering Sciences and Center for 
           Aerospace Structures, University of Colorado. 2004.

    """
    nE, nNE = topo.shape
    diags = np.zeros((nE, 6 * nNE), dtype=dens.dtype)
    for iE in prange(nE):
        li = lengths[iE]
        vi = li * areas[iE]
        di = dens[iE]
        for jNE in prange(nNE):
            diags[iE, jNE*6: jNE*6 + 3] = di * vi / nNE
            diags[iE, jNE*6 + 3: jNE*6 + 6] = di * vi * alpha * li**2
    return diags



# -*- coding: utf-8 -*-
from typing import Union, List, Iterable, Collection
import logging

from scipy.sparse import coo_matrix
import numpy as np
from numpy import ndarray

from neumann import squeeze
from neumann.array import atleast3d

from polymesh import PolyData

from .pointdata import PointData
from .preproc import (nodal_load_vector, essential_penalty_matrix,
                      nodal_mass_matrix)
from .cells import TET4, TET10, H8, H27, FiniteElement
from .metamesh import ABC_FemMesh
from .constants import DEFAULT_DIRICHLET_PENALTY


class FemMesh(PolyData, ABC_FemMesh):
    """
    A descendant of :class:`polymesh.PolyData` to handle polygonal meshes for 
    the **Finite Element Method**.

    Parameters
    ----------
    pd : polymesh.PolyData or polymesh.CellData, Optional
        A PolyData or a CellData instance. Dafault is None.
    cd : polymesh.CellData, Optional
        A CellData instance, if the first argument is provided. Dafault is None.

    Note
    ----
    The suggested way of usage is to create the pointdata and celldata
    objects in advance, and provide them as the firts argument when building 
    the model.

    Examples
    --------
    >>> from sigmaepsilon.solid import FemMesh
    >>> from polymesh.grid import grid
    >>> size = Lx, Ly, Lz = 100, 100, 100
    >>> shape = nx, ny, nz = 10, 10, 10
    >>> coords, topo = grid(size=size, shape=shape, eshape='H27')
    >>> pd = FemMesh(coords=coords)
    >>> pd['A']['Part1'] = FemMesh(topo=topo[:10])
    >>> pd['B']['Part2'] = FemMesh(topo=topo[10:-10])
    >>> pd['C']['Part3'] = FemMesh(topo=topo[-10:])

    See also
    --------
    :class:`polymesh.PolyData`
    :class:`polymesh.CellData`

    """

    dofs = ('UX', 'UY', 'UZ', 'ROTX', 'ROTY', 'ROTZ')
    NDOFN = 6
    _point_class_ = PointData

    def __init__(self, *args, model=None, cell_fields: dict = None,
                 point_fields: dict = None, **kwargs):
        point_fields = {} if point_fields is None else point_fields
        cell_fields = {} if cell_fields is None else cell_fields
        super().__init__(*args, point_fields=point_fields,
                         cell_fields=cell_fields, **kwargs)
        self._model = model

    @property
    def pd(self) -> PointData:
        """
        Returns the attached pointdata.
        
        See Also
        --------
        class:`sigmaepsilon.solid.fem.pointdata.PointData`
        
        """
        return self.pointdata

    @property
    def cd(self) -> FiniteElement:
        """
        Returns the attached celldata.
        
        See Also
        --------
        class:`sigmaepsilon.solid.fem.cells.elem.FiniteElement`
        class:`sigmaepsilon.solid.fem.cells.celldata.CellData`
        
        """
        return self.celldata
    
    def cellblocks(self, *args, **kwargs) -> Collection['FemMesh']:
        """
        Returns an iterable over blocks with cell data.
        """
        return filter(lambda i: i.celldata is not None, 
                      self.blocks(*args, **kwargs))
                    
    def cells_coords(self, *args, points: Iterable = None, 
                     cells: Iterable = None,
                     **kwargs) -> Union[ndarray, List[ndarray]]:
        """
        Returns the coordinates of the cells as a 3d numpy array if the 
        subblocks form a regular mesh, or a list of such arrays.
        """
        if points is None and cells is None:
            return super().cells_coords(*args, **kwargs)
        else:
            blocks = self.cellblocks(inclusive=True)
            kwargs.update(points=points, squeeze=False)
            if cells is not None:
                kwargs.update(cells=cells)
                res = {}
                def foo(b:FemMesh): 
                    return b.cd.coords(*args, **kwargs)
                [res.update(d) for d in map(foo, blocks)]
                return res
            else:
                def foo(b:FemMesh): 
                    return b.cd.coords(*args, **kwargs)
                return np.vstack(list(map(foo, blocks)))

    def element_dof_numbering(self) -> ndarray:
        """
        Returns global ids of the local degrees of freedoms for each cell.
        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b:FemMesh): 
            return b.cd.global_dof_numbering()
        return np.vstack(list(map(foo, blocks)))
    
    def element_fixity(self) -> ndarray:
        """
        Returns element fixity data.
        """
        fixity = []
        for b in self.cellblocks(inclusive=True):
            if b.has_fixity:
                fixity.append(b.cd.fixity.astype(float))
            else:
                nE, nNE, nDOF = len(b.cd), b.cd.NNODE, self.NDOFN
                fixity.append(np.ones(nE, nNE, nDOF))
        return np.vstack(fixity)
    
    def direction_cosine_matrix(self, target:str='global'):
        blocks = self.cellblocks(inclusive=True)
        def foo(b:FemMesh): 
            return b.cd.direction_cosine_matrix(target=target)
        return np.vstack(list(map(foo, blocks)))

    def elastic_stiffness_matrix(self, *, 
                                 eliminate_zeros: bool = True,
                                 sum_duplicates: bool = True, 
                                 sparse: bool = False,
                                 transform:bool = True,
                                 **kwargs) -> Union[ndarray, coo_matrix]:
        """
        Returns the elastic stiffness matrix in dense or sparse format.

        Parameters
        ----------
        sparse : bool, Optional
            If True, the result is a sparse matrix. Default is False.
        eliminate_zeros : bool, Optional
            Eliminates zero entries. Only if 'sparse' is True. Default is True.
        sum_duplicates : bool, Optional
            Sums duplicate entries. Only if 'sparse' is True. Default is True.
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.
            
        Returns
        -------
        numpy.ndarray or scipy.sparse.coo_matrix

        """
        blocks = self.cellblocks(inclusive=True)
        if sparse:
            assert transform, "Must transform for sparse layout."
            def foo(b:FemMesh): 
                return b.cd.elastic_stiffness_matrix(sparse=True, transform=True)
            K = np.sum(list(map(foo, blocks))).tocoo()
            if eliminate_zeros:
                K.eliminate_zeros()
            if sum_duplicates:
                K.sum_duplicates()
            return K
        else:
            def foo(b:FemMesh): 
                return b.cd.elastic_stiffness_matrix(transform=transform)
            return np.vstack(list(map(foo, blocks)))

    def masses(self, *args, **kwargs) -> ndarray:
        """
        Returns masses of the cells as a 1d numpy array.
        """
        blocks = self.cellblocks(*args, inclusive=True, **kwargs)
        def foo(b:FemMesh): 
            return b.cd.masses()
        vmap = map(foo, blocks)
        return np.concatenate(list(vmap))

    def mass(self, *args, **kwargs) -> float:
        """
        Returns the total mass of the structure.
        """
        return np.sum(self.masses(*args, **kwargs))

    def consistent_mass_matrix(self, *, eliminate_zeros: bool = True,
                               sum_duplicates: bool = True, sparse: bool = False,
                               transform:bool = True,**kwargs) \
                                   -> Union[ndarray, coo_matrix]:
        """
        Returns the stiffness-consistent mass matrix as a dense or a sparse 
        matrix.

        Parameters
        ----------
        sparse : bool, Optional
            If True, the result is a sparse matrix. Default is False.
        eliminate_zeros : bool, Optional
            Eliminates zero entries. Only if 'sparse' is True. Default is True.
        sum_duplicates : bool, Optional
            Sums duplicate entries. Only if 'sparse' is True. Default is True.
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is True.
            
        Returns
        -------
        numpy.ndarray or scipy.sparse.coo_matrix
            The mass matrix as a 3d dense or a 2d sparse array.

        """
        # distributed masses (cells)
        blocks = self.cellblocks(inclusive=True)
        if sparse:
            def foo(b:FemMesh): 
                return b.cd.consistent_mass_matrix(sparse=True, transform=True)
            M = np.sum(list(map(foo, blocks))).tocoo()
            if eliminate_zeros:
                M.eliminate_zeros()
            if sum_duplicates:
                M.sum_duplicates()
            return M
        else:
            def foo(b:FemMesh): 
                return b.cd.consistent_mass_matrix(transform=transform)
            return np.vstack(list(map(foo, blocks)))

    def mass_matrix(self, *, eliminate_zeros: bool = True,
                    sum_duplicates: bool = True, **kwargs) -> coo_matrix:
        """
        Returns the mass matrix in coo format.The resulting matrix is the 
        sum of the consistent mass matrix and the nodal mass matrix.

        Parameters
        ----------
        eliminate_zeros : bool, Optional
            Eliminates zero entries. Default is True.
        sum_duplicates : bool, Optional
            Sums duplicate entries. Default is True.

        Returns
        -------
        scipy.sparse.coo_matrix
            The mass matrix in sparse COO format.

        """
        # distributed masses (cells)
        M = self.consistent_mass_matrix(sparse=True,
                                        transform=True,
                                        eliminate_zeros=False,
                                        sum_duplicates=False)
        # nodal masses
        M += self.nodal_mass_matrix(eliminate_zeros=False,
                                    sum_duplicates=False)
        if eliminate_zeros:
            M.eliminate_zeros()
        if sum_duplicates:
            M.sum_duplicates()
        return M.tocoo()

    def nodal_mass_matrix(self, *, eliminate_zeros: bool = False,
                          sum_duplicates: bool = False, **kw):
        """
        Returns the mass matrix from nodal masses only, in coo format.

        Parameters
        ----------
        eliminate_zeros : bool, Optional
            Eliminates zero entries. Default is True.
        sum_duplicates : bool, Optional
            Sums duplicate entries. Default is True.

        Returns
        -------
        scipy.sparse.coo_matrix
            The nodal mass matrix in sparse COO format.

        """
        pd = self.root().pointdata
        nN = len(pd)
        nDOF = self.__class__.NDOFN
        try:
            m = pd.mass
            M = nodal_mass_matrix(values=m, eliminate_zeros=eliminate_zeros,
                                  sum_duplicates=sum_duplicates, ndof=nDOF)
            return M
        except Exception as e:
            nX = nN * nDOF
            return coo_matrix(shape=(nX, nX))

    def essential_penalty_matrix(self, fixity: ndarray = None, *,
                                 eliminate_zeros: bool = True,
                                 sum_duplicates: bool = True,
                                 penalty: float = DEFAULT_DIRICHLET_PENALTY,
                                 **kwargs) -> coo_matrix:
        """
        A penalty matrix that enforces Dirichlet boundary conditions. 
        Returns a scipy sparse matrix in coo format.

        Parameters
        ----------
        fixity : numpy.ndarray, Optional
            The fixity values as a 2d array. The length of the first axis must 
            equal the number of nodes, the second the number of DOFs of the model.
            If not specified, the data is assumed to be stored in the attached 
            pointdata with the same key. Default is None.
        penalty : float, Optional
            The Courant-type penalty value.
            Default is `sigmaepsilon.solid.fem.constants.DEFAULT_DIRICHLET_PENALTY`.
        eliminate_zeros : bool, Optional
            Eliminates zeros from the matrix. Default is True.
        eliminate_zeros : bool, Optional
            Summs duplicate entries in the matrix. Default is True.

        Returns
        -------
        scipy.sparse.coo_matrix

        """
        root = self.root()
        pd = root.pointdata
        N = len(pd) * root.NDOFN
        K_coo = coo_matrix(([0.], ([0], [0])), shape=(N, N))
        try:
            fixity = pd.fixity
            K_coo += essential_penalty_matrix(fixity=fixity, pfix=penalty,
                                              eliminate_zeros=eliminate_zeros,
                                              sum_duplicates=sum_duplicates)
        except AttributeError:
            logging.info('No FIXITY information in pointdata.')
        except Exception as e:
            logging.error(f'Exception {e} occured', exc_info=True)
        return K_coo.tocoo()

    @squeeze(True)
    def nodal_load_vector(self, *args, **kwargs) -> ndarray:
        """
        Returns the nodal load vector.

        Returns
        -------
        numpy.ndarray

        """
        # concentrated nodal loads
        nodal_data = self.root().pointdata.loads
        nodal_data = atleast3d(nodal_data, back=True)
        # (nP, nDOF, nRHS)
        res = nodal_load_vector(nodal_data, squeeze=False)
        return res
    
    @squeeze(True)
    def cell_load_vector(self, *, transform: bool = True, 
                         assemble: bool = False, **kwargs) -> ndarray:
        """
        Returns the nodal load vector from body and strain loads.
        
        Parameters
        ----------
        assemble : bool, Optional
            If True, the values are returned with a matching shape to 
            the total system. Default is False.
        transform : bool, Optional
            If True, local matrices are transformed to the global frame.
            Default is False.       

        Returns
        -------
        numpy.ndarray
            The shape depends on the arguments.
            
        """
        if assemble:
            assert transform, "Must transform before assembly."
        blocks = list(self.cellblocks(inclusive=True))
        params = dict(
            squeeze=False,
            transform=transform,
            assemble=assemble
        )
        params.update(**kwargs)
        def foo(b:FemMesh): 
            return b.cd.load_vector(**params)
        return np.vstack(list(map(foo, blocks)))

    def prostproc_dof_solution(self, *args, **kwargs):
        """
        Calculates approximate solution of the primary variables as a list 
        of arrays for each block in the mesh.
        """
        blocks = self.cellblocks(inclusive=True)
        dofsol = self.root().pointdata.dofsol
        def foo(b:FemMesh): 
            return b.cd.prostproc_dof_solution(dofsol=dofsol)
        list(map(foo, blocks))
        
    def condensate_cell_fixity(self) -> 'FemMesh':
        """
        Performs static condensation of the system equations to account for
        cell fixity. Returns the mesh object for continuation.
        """
        [b.cd.condensate() for b in self.cellblocks(inclusive=True)]
        return self

    @squeeze(True)
    def nodal_dof_solution(self, *, flatten: bool = False, **kw) -> ndarray:
        """
        Returns nodal degree of freedom solution.

        Returns
        -------
        numpy.ndarray

        """
        dofsol = self.root().pointdata.dofsol
        if flatten:
            if len(dofsol.shape) == 2:
                return dofsol.flatten()
            else:
                nN, nDOFN, nRHS = dofsol.shape
                return dofsol.reshape((nN * nDOFN, nRHS))
        else:
            return dofsol

    @squeeze(True)
    def cell_dof_solution(self, *args, cells:Iterable=None, 
                          flatten: bool = True, squeeze: bool = True, 
                          **kwargs) -> ndarray:
        """
        Returns degree of freedom solution for each cell.

        Returns
        -------
        numpy.ndarray

        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(flatten=flatten, squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b:FemMesh): 
                return b.cd.dof_solution(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b:FemMesh): 
                return b.cd.dof_solution(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    @squeeze(True)
    def strains(self, *args, cells:Iterable=None, squeeze: bool = True,
                **kwargs) -> ndarray:
        """
        Returns strains for each cell.

        Returns
        -------
        numpy.ndarray

        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b:FemMesh): 
                return b.cd.strains(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b:FemMesh): 
                return b.cd.strains(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    @squeeze(True)
    def external_forces(self, *args, cells:Iterable=None, 
                        flatten: bool = True, squeeze: bool = True, 
                        **kwargs) -> ndarray:
        """
        Returns external forces for each cell.

        Returns
        -------
        numpy.ndarray

        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(flatten=flatten, squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b:FemMesh): 
                return b.cd.external_forces(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b:FemMesh): 
                return b.cd.external_forces(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    @squeeze(True)
    def internal_forces(self, *args, cells:Iterable=None, flatten: bool = True,
                        squeeze: bool = True, **kwargs) -> ndarray:
        """
        Returns internal forces for each cell.

        Returns
        -------
        numpy.ndarray

        """
        blocks = self.cellblocks(inclusive=True)
        kwargs.update(flatten=flatten, squeeze=False)
        if cells is not None:
            kwargs.update(cells=cells)
            res = {}
            def foo(b:FemMesh): 
                return b.cd.internal_forces(*args, **kwargs)
            [res.update(d) for d in map(foo, blocks)]
            return res
        else:
            def foo(b:FemMesh): 
                return b.cd.internal_forces(*args, **kwargs)
            return np.vstack(list(map(foo, blocks)))

    def stresses_at_cells_nodes(self, *args, **kwargs) -> ndarray:
        """
        Returns stresses at all nodes of all cells.
        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b:FemMesh): 
            return b.cd.stresses_at_nodes(*args, **kwargs)
        return np.vstack(list(map(foo, blocks)))

    def stresses_at_centers(self, *args, **kwargs) -> ndarray:
        """
        Returns stresses at the centers of all cells.

        Returns
        -------
        numpy.ndarray

        """
        blocks = self.cellblocks(inclusive=True)
        def foo(b:FemMesh): 
            return b.cd.stresses_at_centers(*args, **kwargs)
        return np.squeeze(np.vstack(list(map(foo, blocks))))

    @property
    def model(self):
        """
        Returns the attached model object if there is any.
        """
        if self._model is not None:
            return self._model
        else:
            if self.is_root():
                return self._model
            else:
                return self.parent.model

    def material_stiffness_matrix(self) -> ndarray:
        """
        Returns the model of the mesh if there is any.
        This is not important if all the celldata in the mesh
        are filled up with appropriate model stiffness matrices.
        Otherwise, the model stiffness matrix is calcualated on the 
        mesh level during the initialization stage and is spread 
        among the subblocks.
        """
        m = self.model
        if isinstance(m, np.ndarray):
            return m
        else:
            try:
                return m.stiffness_matrix()
            except Exception:
                raise RuntimeError("Invalid model type {}".format(type(m)))

    def postprocess(self, *args, **kwargs):
        """
        General postprocessing. This may be reimplemented in subclasses,
        currently nothing happens here.
        """
        pass


class SolidMesh(FemMesh):
    """
    A data class dedicated to simple 3d solid cells with 3 degrees of 
    freedom per node.    

    See Also
    --------
    :class:`TET4`
    :class:`TET10`
    :class:`H8`
    :class:`H27`

    """

    dofs = ('UX', 'UY', 'UZ')
    NDOFN = 3

    _cell_classes_ = {
        4: TET4,
        10: TET10,
        8: H8,
        27: H27,
    }
